from __future__ import annotations

import datetime
from datetime import timezone

from hcloud.certificates import Certificate


class TestCertificate:
    def test_created_is_datetime(self):
        certificate = Certificate(
            id=1,
            created="2016-01-30T23:50+00:00",
            not_valid_after="2016-01-30T23:50+00:00",
            not_valid_before="2016-01-30T23:50+00:00",
        )
        assert certificate.created == datetime.datetime(
            2016, 1, 30, 23, 50, tzinfo=timezone.utc
        )
        assert certificate.not_valid_after == datetime.datetime(
            2016, 1, 30, 23, 50, tzinfo=timezone.utc
        )
        assert certificate.not_valid_before == datetime.datetime(
            2016, 1, 30, 23, 50, tzinfo=timezone.utc
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from dateutil.parser import isoparse

from ..core import BaseDomain

if TYPE_CHECKING:
    from ..actions import BoundAction
    from ..locations import BoundLocation
    from ..servers import BoundServer
    from .client import BoundFloatingIP


class FloatingIP(BaseDomain):
    """Floating IP Domain

    :param id: int
           ID of the Floating IP
    :param description: str, None
           Description of the Floating IP
    :param ip: str
           IP address of the Floating IP
    :param type: str
           Type of Floating IP. Choices: `ipv4`, `ipv6`
    :param server: :class:`BoundServer <hcloud.servers.client.BoundServer>`, None
           Server the Floating IP is assigned to, None if it is not assigned at all
    :param dns_ptr: List[Dict]
           Array of reverse DNS entries
    :param home_location: :class:`BoundLocation <hcloud.locations.client.BoundLocation>`
           Location the Floating IP was created in. Routing is optimized for this location.
    :param blocked: boolean
           Whether the IP is blocked
    :param protection: dict
           Protection configuration for the Floating IP
    :param labels: dict
           User-defined labels (key-value pairs)
    :param created: datetime
           Point in time when the Floating IP was created
    :param name: str
           Name of the Floating IP
    """

    __slots__ = (
        "id",
        "type",
        "description",
        "ip",
        "server",
        "dns_ptr",
        "home_location",
        "blocked",
        "protection",
        "labels",
        "name",
        "created",
    )

    def __init__(
        self,
        id: int | None = None,
        type: str | None = None,
        description: str | None = None,
        ip: str | None = None,
        server: BoundServer | None = None,
        dns_ptr: list[dict] | None = None,
        home_location: BoundLocation | None = None,
        blocked: bool | None = None,
        protection: dict | None = None,
        labels: dict[str, str] | None = None,
        created: str | None = None,
        name: str | None = None,
    ):
        self.id = id
        self.type = type
        self.description = description
        self.ip = ip
        self.server = server
        self.dns_ptr = dns_ptr
        self.home_location = home_location
        self.blocked = blocked
        self.protection = protection
        self.labels = labels
        self.created = isoparse(created) if created else None
        self.name = name


class CreateFloatingIPResponse(BaseDomain):
    """Create Floating IP Response Domain

    :param floating_ip: :class:`BoundFloatingIP <hcloud.floating_ips.client.BoundFloatingIP>`
           The Floating IP which was created
    :param action: :class:`BoundAction <hcloud.actions.client.BoundAction>`
           The Action which shows the progress of the Floating IP Creation
    """

    __slots__ = ("floating_ip", "action")

    def __init__(
        self,
        floating_ip: BoundFloatingIP,
        action: BoundAction | None,
    ):
        self.floating_ip = floating_ip
        self.action = action



from __future__ import annotations

from typing import TYPE_CHECKING

from dateutil.parser import isoparse

from ..core import BaseDomain, DomainIdentityMixin

if TYPE_CHECKING:
    from ..actions import BoundAction
    from ..locations import BoundLocation, Location
    from ..servers import BoundServer, Server
    from .client import BoundVolume


class Volume(BaseDomain, DomainIdentityMixin):
    """Volume Domain

    :param id: int
           ID of the Volume
    :param name: str
           Name of the Volume
    :param server: :class:`BoundServer <hcloud.servers.client.BoundServer>`, None
           Server the Volume is attached to, None if it is not attached at all.
    :param created: datetime
           Point in time when the Volume was created
    :param location: :class:`BoundLocation <hcloud.locations.client.BoundLocation>`
           Location of the Volume. Volume can only be attached to Servers in the same location.
    :param size: int
           Size in GB of the Volume
    :param linux_device: str
           Device path on the file system for the Volume
    :param protection: dict
           Protection configuration for the Volume
    :param labels: dict
           User-defined labels (key-value pairs)
    :param status: str
           Current status of the volume Choices: `creating`, `available`
    :param format: str, None
           Filesystem of the volume if formatted on creation, None if not formatted on creation.
    """

    STATUS_CREATING = "creating"
    """Volume Status creating"""
    STATUS_AVAILABLE = "available"
    """Volume Status available"""

    __slots__ = (
        "id",
        "name",
        "server",
        "location",
        "size",
        "linux_device",
        "format",
        "protection",
        "labels",
        "status",
        "created",
    )

    def __init__(
        self,
        id: int,
        name: str | None = None,
        server: Server | BoundServer | None = None,
        created: str | None = None,
        location: Location | BoundLocation | None = None,
        size: int | None = None,
        linux_device: str | None = None,
        format: str | None = None,
        protection: dict | None = None,
        labels: dict[str, str] | None = None,
        status: str | None = None,
    ):
        self.id = id
        self.name = name
        self.server = server
        self.created = isoparse(created) if created else None
        self.location = location
        self.size = size
        self.linux_device = linux_device
        self.format = format
        self.protection = protection
        self.labels = labels
        self.status = status


class CreateVolumeResponse(BaseDomain):
    """Create Volume Response Domain

    :param volume: :class:`BoundVolume <hcloud.volumes.client.BoundVolume>`
           The created volume
    :param action: :class:`BoundAction <hcloud.actions.client.BoundAction>`
           The action that shows the progress of the Volume Creation
    :param next_actions: List[:class:`BoundAction <hcloud.actions.client.BoundAction>`]
           List of actions that are performed after the creation, like attaching to a server
    """

    __slots__ = ("volume", "action", "next_actions")

    def __init__(
        self,
        volume: BoundVolume,
        action: BoundAction,
        next_actions: list[BoundAction],
    ):
        self.volume = volume
        self.action = action
        self.next_actions = next_actions



from __future__ import annotations

from .client import BoundServer, ServersClient, ServersPageResult  # noqa: F401
from .domain import (  # noqa: F401
    CreateServerResponse,
    EnableRescueResponse,
    GetMetricsResponse,
    IPv4Address,
    IPv6Network,
    PrivateNet,
    PublicNetwork,
    PublicNetworkFirewall,
    RequestConsoleResponse,
    ResetPasswordResponse,
    Server,
    ServerCreatePublicNetwork,
)



from __future__ import annotations

from unittest import mock

import pytest

from hcloud.load_balancer_types import LoadBalancerTypesClient


class TestLoadBalancerTypesClient:
    @pytest.fixture()
    def load_balancer_types_client(self):
        return LoadBalancerTypesClient(client=mock.MagicMock())

    def test_get_by_id(self, load_balancer_types_client, load_balancer_type_response):
        load_balancer_types_client._client.request.return_value = (
            load_balancer_type_response
        )
        load_balancer_type = load_balancer_types_client.get_by_id(1)
        load_balancer_types_client._client.request.assert_called_with(
            url="/load_balancer_types/1", method="GET"
        )
        assert load_balancer_type._client is load_balancer_types_client
        assert load_balancer_type.id == 1
        assert load_balancer_type.name == "LB11"

    @pytest.mark.parametrize(
        "params", [{"name": "LB11", "page": 1, "per_page": 10}, {"name": ""}, {}]
    )
    def test_get_list(
        self, load_balancer_types_client, two_load_balancer_types_response, params
    ):
        load_balancer_types_client._client.request.return_value = (
            two_load_balancer_types_response
        )
        result = load_balancer_types_client.get_list(**params)
        load_balancer_types_client._client.request.assert_called_with(
            url="/load_balancer_types", method="GET", params=params
        )

        load_balancer_types = result.load_balancer_types
        assert result.meta is None

        assert len(load_balancer_types) == 2

        load_balancer_types1 = load_balancer_types[0]
        load_balancer_types2 = load_balancer_types[1]

        assert load_balancer_types1._client is load_balancer_types_client
        assert load_balancer_types1.id == 1
        assert load_balancer_types1.name == "LB11"

        assert load_balancer_types2._client is load_balancer_types_client
        assert load_balancer_types2.id == 2
        assert load_balancer_types2.name == "LB21"

    @pytest.mark.parametrize("params", [{"name": "LB21"}])
    def test_get_all(
        self, load_balancer_types_client, two_load_balancer_types_response, params
    ):
        load_balancer_types_client._client.request.return_value = (
            two_load_balancer_types_response
        )
        load_balancer_types = load_balancer_types_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        load_balancer_types_client._client.request.assert_called_with(
            url="/load_balancer_types", method="GET", params=params
        )

        assert len(load_balancer_types) == 2

        load_balancer_types1 = load_balancer_types[0]
        load_balancer_types2 = load_balancer_types[1]

        assert load_balancer_types1._client is load_balancer_types_client
        assert load_balancer_types1.id == 1
        assert load_balancer_types1.name == "LB11"

        assert load_balancer_types2._client is load_balancer_types_client
        assert load_balancer_types2.id == 2
        assert load_balancer_types2.name == "LB21"

    def test_get_by_name(
        self, load_balancer_types_client, one_load_balancer_types_response
    ):
        load_balancer_types_client._client.request.return_value = (
            one_load_balancer_types_response
        )
        load_balancer_type = load_balancer_types_client.get_by_name("LB21")

        params = {"name": "LB21"}

        load_balancer_types_client._client.request.assert_called_with(
            url="/load_balancer_types", method="GET", params=params
        )

        assert load_balancer_type._client is load_balancer_types_client
        assert load_balancer_type.id == 2
        assert load_balancer_type.name == "LB21"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

from ..core import BoundModelBase, ClientEntityBase, Meta
from .domain import LoadBalancerType

if TYPE_CHECKING:
    from .._client import Client


class BoundLoadBalancerType(BoundModelBase, LoadBalancerType):
    _client: LoadBalancerTypesClient

    model = LoadBalancerType


class LoadBalancerTypesPageResult(NamedTuple):
    load_balancer_types: list[BoundLoadBalancerType]
    meta: Meta | None


class LoadBalancerTypesClient(ClientEntityBase):
    _client: Client

    def get_by_id(self, id: int) -> BoundLoadBalancerType:
        """Returns a specific Load Balancer Type.

        :param id: int
        :return: :class:`BoundLoadBalancerType <hcloud.load_balancer_type.client.BoundLoadBalancerType>`
        """
        response = self._client.request(
            url=f"/load_balancer_types/{id}",
            method="GET",
        )
        return BoundLoadBalancerType(self, response["load_balancer_type"])

    def get_list(
        self,
        name: str | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> LoadBalancerTypesPageResult:
        """Get a list of Load Balancer types

        :param name: str (optional)
               Can be used to filter Load Balancer type by their name.
        :param page: int (optional)
               Specifies the page to fetch
        :param per_page: int (optional)
               Specifies how many results are returned by page
        :return: (List[:class:`BoundLoadBalancerType <hcloud.load_balancer_types.client.BoundLoadBalancerType>`], :class:`Meta <hcloud.core.domain.Meta>`)
        """
        params: dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page

        response = self._client.request(
            url="/load_balancer_types", method="GET", params=params
        )
        load_balancer_types = [
            BoundLoadBalancerType(self, load_balancer_type_data)
            for load_balancer_type_data in response["load_balancer_types"]
        ]
        return LoadBalancerTypesPageResult(
            load_balancer_types, Meta.parse_meta(response)
        )

    def get_all(self, name: str | None = None) -> list[BoundLoadBalancerType]:
        """Get all Load Balancer types

        :param name: str (optional)
               Can be used to filter Load Balancer type by their name.
        :return: List[:class:`BoundLoadBalancerType <hcloud.load_balancer_types.client.BoundLoadBalancerType>`]
        """
        return self._iter_pages(self.get_list, name=name)

    def get_by_name(self, name: str) -> BoundLoadBalancerType | None:
        """Get Load Balancer type by name

        :param name: str
               Used to get Load Balancer type by name.
        :return: :class:`BoundLoadBalancerType <hcloud.load_balancer_types.client.BoundLoadBalancerType>`
        """
        return self._get_first_by(name=name)



from __future__ import annotations

from ..core import BaseDomain, DomainIdentityMixin


class Location(BaseDomain, DomainIdentityMixin):
    """Location Domain

    :param id: int
           ID of location
    :param name: str
           Name of location
    :param description: str
           Description of location
    :param country: str
           ISO 3166-1 alpha-2 code of the country the location resides in
    :param city: str
           City the location is closest to
    :param latitude: float
           Latitude of the city closest to the location
    :param longitude: float
           Longitude of the city closest to the location
    :param network_zone: str
           Name of network zone this location resides in
    """

    __slots__ = (
        "id",
        "name",
        "description",
        "country",
        "city",
        "latitude",
        "longitude",
        "network_zone",
    )

    def __init__(
        self,
        id: int | None = None,
        name: str | None = None,
        description: str | None = None,
        country: str | None = None,
        city: str | None = None,
        latitude: float | None = None,
        longitude: float | None = None,
        network_zone: str | None = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.country = country
        self.city = city
        self.latitude = latitude
        self.longitude = longitude
        self.network_zone = network_zone



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

from ..actions import ActionsPageResult, BoundAction, ResourceActionsClient
from ..core import BoundModelBase, ClientEntityBase, Meta
from .domain import Image

if TYPE_CHECKING:
    from .._client import Client


class BoundImage(BoundModelBase, Image):
    _client: ImagesClient

    model = Image

    def __init__(self, client: ImagesClient, data: dict):
        # pylint: disable=import-outside-toplevel
        from ..servers import BoundServer

        created_from = data.get("created_from")
        if created_from is not None:
            data["created_from"] = BoundServer(
                client._client.servers, created_from, complete=False
            )
        bound_to = data.get("bound_to")
        if bound_to is not None:
            data["bound_to"] = BoundServer(
                client._client.servers, {"id": bound_to}, complete=False
            )

        super().__init__(client, data)

    def get_actions_list(
        self,
        sort: list[str] | None = None,
        page: int | None = None,
        per_page: int | None = None,
        status: list[str] | None = None,
    ) -> ActionsPageResult:
        """Returns a list of action objects for the image.

        :param status: List[str] (optional)
               Response will have only actions with specified statuses. Choices: `running` `success` `error`
        :param sort: List[str] (optional)
               Specify how the results are sorted. Choices: `id` `id:asc` `id:desc` `command` `command:asc` `command:desc` `status` `status:asc` `status:desc` `progress` `progress:asc` `progress:desc` `started` `started:asc` `started:desc` `finished` `finished:asc` `finished:desc`
        :param page: int (optional)
               Specifies the page to fetch
        :param per_page: int (optional)
               Specifies how many results are returned by page
        :return: (List[:class:`BoundAction <hcloud.actions.client.BoundAction>`], :class:`Meta <hcloud.core.domain.Meta>`)
        """
        return self._client.get_actions_list(
            self, sort=sort, page=page, per_page=per_page, status=status
        )

    def get_actions(
        self,
        sort: list[str] | None = None,
        status: list[str] | None = None,
    ) -> list[BoundAction]:
        """Returns all action objects for the image.

        :param status: List[str] (optional)
               Response will have only actions with specified statuses. Choices: `running` `success` `error`
        :param sort: List[str] (optional)
               Specify how the results are sorted. Choices: `id` `id:asc` `id:desc` `command` `command:asc` `command:desc` `status` `status:asc` `status:desc` `progress` `progress:asc` `progress:desc` `started` `started:asc` `started:desc` `finished` `finished:asc` `finished:desc`
        :return: List[:class:`BoundAction <hcloud.actions.client.BoundAction>`]
        """
        return self._client.get_actions(self, status=status, sort=sort)

    def update(
        self,
        description: str | None = None,
        type: str | None = None,
        labels: dict[str, str] | None = None,
    ) -> BoundImage:
        """Updates the Image. You may change the description, convert a Backup image to a Snapshot Image or change the image labels.

        :param description: str (optional)
               New description of Image
        :param type: str (optional)
               Destination image type to convert to
               Choices: snapshot
        :param labels: Dict[str, str] (optional)
               User-defined labels (key-value pairs)
        :return: :class:`BoundImage <hcloud.images.client.BoundImage>`
        """
        return self._client.update(self, description, type, labels)

    def delete(self) -> bool:
        """Deletes an Image. Only images of type snapshot and backup can be deleted.

        :return: bool
        """
        return self._client.delete(self)

    def change_protection(self, delete: bool | None = None) -> BoundAction:
        """Changes the protection configuration of the image. Can only be used on snapshots.

        :param delete: bool
               If true, prevents the snapshot from being deleted
        :return: :class:`BoundAction <hcloud.actions.client.BoundAction>`
        """
        return self._client.change_protection(self, delete)


class ImagesPageResult(NamedTuple):
    images: list[BoundImage]
    meta: Meta | None


class ImagesClient(ClientEntityBase):
    _client: Client

    actions: ResourceActionsClient
    """Images scoped actions client

    :type: :class:`ResourceActionsClient <hcloud.actions.client.ResourceActionsClient>`
    """

    def __init__(self, client: Client):
        super().__init__(client)
        self.actions = ResourceActionsClient(client, "/images")

    def get_actions_list(
        self,
        image: Image | BoundImage,
        sort: list[str] | None = None,
        page: int | None = None,
        per_page: int | None = None,
        status: list[str] | None = None,
    ) -> ActionsPageResult:
        """Returns a list of action objects for an image.

        :param image: :class:`BoundImage <hcloud.images.client.BoundImage>` or :class:`Image <hcloud.images.domain.Image>`
        :param status: List[str] (optional)
               Response will have only actions with specified statuses. Choices: `running` `success` `error`
        :param sort: List[str] (optional)
               Specify how the results are sorted. Choices: `id` `id:asc` `id:desc` `command` `command:asc` `command:desc` `status` `status:asc` `status:desc` `progress` `progress:asc` `progress:desc` `started` `started:asc` `started:desc` `finished` `finished:asc` `finished:desc`
        :param page: int (optional)
               Specifies the page to fetch
        :param per_page: int (optional)
               Specifies how many results are returned by page
        :return: (List[:class:`BoundAction <hcloud.actions.client.BoundAction>`], :class:`Meta <hcloud.core.domain.Meta>`)
        """
        params: dict[str, Any] = {}
        if sort is not None:
            params["sort"] = sort
        if status is not None:
            params["status"] = status
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page
        response = self._client.request(
            url=f"/images/{image.id}/actions",
            method="GET",
            params=params,
        )
        actions = [
            BoundAction(self._client.actions, action_data)
            for action_data in response["actions"]
        ]
        return ActionsPageResult(actions, Meta.parse_meta(response))

    def get_actions(
        self,
        image: Image | BoundImage,
        sort: list[str] | None = None,
        status: list[str] | None = None,
    ) -> list[BoundAction]:
        """Returns all action objects for an image.

        :param image: :class:`BoundImage <hcloud.images.client.BoundImage>` or :class:`Image <hcloud.images.domain.Image>`
        :param status: List[str] (optional)
               Response will have only actions with specified statuses. Choices: `running` `success` `error`
        :param sort: List[str] (optional)
               Specify how the results are sorted. Choices: `id` `command` `status` `progress`  `started` `finished` . You can add one of ":asc", ":desc" to modify sort order. ( ":asc" is default)
        :return: List[:class:`BoundAction <hcloud.actions.client.BoundAction>`]
        """
        return self._iter_pages(
            self.get_actions_list,
            image,
            sort=sort,
            status=status,
        )

    def get_by_id(self, id: int) -> BoundImage:
        """Get a specific Image

        :param id: int
        :return: :class:`BoundImage <hcloud.images.client.BoundImage`
        """
        response = self._client.request(url=f"/images/{id}", method="GET")
        return BoundImage(self, response["image"])

    def get_list(
        self,
        name: str | None = None,
        label_selector: str | None = None,
        bound_to: list[str] | None = None,
        type: list[str] | None = None,
        architecture: list[str] | None = None,
        sort: list[str] | None = None,
        page: int | None = None,
        per_page: int | None = None,
        status: list[str] | None = None,
        include_deprecated: bool | None = None,
    ) -> ImagesPageResult:
        """Get all images

        :param name: str (optional)
               Can be used to filter images by their name.
        :param label_selector: str (optional)
               Can be used to filter servers by labels. The response will only contain servers matching the label selector.
        :param bound_to: List[str] (optional)
               Server Id linked to the image. Only available for images of type backup
        :param type: List[str] (optional)
               Choices: system snapshot backup
        :param architecture: List[str] (optional)
               Choices: x86 arm
        :param status: List[str] (optional)
               Can be used to filter images by their status. The response will only contain images matching the status.
        :param sort: List[str] (optional)
               Choices: id id:asc id:desc name name:asc name:desc created created:asc created:desc
        :param include_deprecated: bool (optional)
               Include deprecated images in the response. Default: False
        :param page: int (optional)
               Specifies the page to fetch
        :param per_page: int (optional)
               Specifies how many results are returned by page
        :return: (List[:class:`BoundImage <hcloud.images.client.BoundImage>`], :class:`Meta <hcloud.core.domain.Meta>`)
        """
        params: dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if label_selector is not None:
            params["label_selector"] = label_selector
        if bound_to is not None:
            params["bound_to"] = bound_to
        if type is not None:
            params["type"] = type
        if architecture is not None:
            params["architecture"] = architecture
        if sort is not None:
            params["sort"] = sort
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page
        if status is not None:
            params["status"] = per_page
        if include_deprecated is not None:
            params["include_deprecated"] = include_deprecated
        response = self._client.request(url="/images", method="GET", params=params)
        images = [BoundImage(self, image_data) for image_data in response["images"]]

        return ImagesPageResult(images, Meta.parse_meta(response))

    def get_all(
        self,
        name: str | None = None,
        label_selector: str | None = None,
        bound_to: list[str] | None = None,
        type: list[str] | None = None,
        architecture: list[str] | None = None,
        sort: list[str] | None = None,
        status: list[str] | None = None,
        include_deprecated: bool | None = None,
    ) -> list[BoundImage]:
        """Get all images

        :param name: str (optional)
               Can be used to filter images by their name.
        :param label_selector: str (optional)
               Can be used to filter servers by labels. The response will only contain servers matching the label selector.
        :param bound_to: List[str] (optional)
               Server Id linked to the image. Only available for images of type backup
        :param type: List[str] (optional)
               Choices: system snapshot backup
        :param architecture: List[str] (optional)
               Choices: x86 arm
        :param status: List[str] (optional)
               Can be used to filter images by their status. The response will only contain images matching the status.
        :param sort: List[str] (optional)
               Choices: id name created (You can add one of ":asc", ":desc" to modify sort order. ( ":asc" is default))
        :param include_deprecated: bool (optional)
               Include deprecated images in the response. Default: False
        :return: List[:class:`BoundImage <hcloud.images.client.BoundImage>`]
        """
        return self._iter_pages(
            self.get_list,
            name=name,
            label_selector=label_selector,
            bound_to=bound_to,
            type=type,
            architecture=architecture,
            sort=sort,
            status=status,
            include_deprecated=include_deprecated,
        )

    def get_by_name(self, name: str) -> BoundImage | None:
        """Get image by name

        Deprecated: Use get_by_name_and_architecture instead.

        :param name: str
               Used to get image by name.
        :return: :class:`BoundImage <hcloud.images.client.BoundImage>`
        """
        return self._get_first_by(name=name)

    def get_by_name_and_architecture(
        self,
        name: str,
        architecture: str,
    ) -> BoundImage | None:
        """Get image by name

        :param name: str
               Used to identify the image.
        :param architecture: str
               Used to identify the image.
        :return: :class:`BoundImage <hcloud.images.client.BoundImage>`
        """
        return self._get_first_by(name=name, architecture=[architecture])

    def update(
        self,
        image: Image | BoundImage,
        description: str | None = None,
        type: str | None = None,
        labels: dict[str, str] | None = None,
    ) -> BoundImage:
        """Updates the Image. You may change the description, convert a Backup image to a Snapshot Image or change the image labels.

        :param image: :class:`BoundImage <hcloud.images.client.BoundImage>` or :class:`Image <hcloud.images.domain.Image>`
        :param description: str (optional)
               New description of Image
        :param type: str (optional)
               Destination image type to convert to
               Choices: snapshot
        :param labels: Dict[str, str] (optional)
               User-defined labels (key-value pairs)
        :return: :class:`BoundImage <hcloud.images.client.BoundImage>`
        """
        data: dict[str, Any] = {}
        if description is not None:
            data.update({"description": description})
        if type is not None:
            data.update({"type": type})
        if labels is not None:
            data.update({"labels": labels})
        response = self._client.request(
            url=f"/images/{image.id}", method="PUT", json=data
        )
        return BoundImage(self, response["image"])

    def delete(self, image: Image | BoundImage) -> bool:
        """Deletes an Image. Only images of type snapshot and backup can be deleted.

        :param :class:`BoundImage <hcloud.images.client.BoundImage>` or :class:`Image <hcloud.images.domain.Image>`
        :return: bool
        """
        self._client.request(url=f"/images/{image.id}", method="DELETE")
        # Return allays true, because the API does not return an action for it. When an error occurs a APIException will be raised
        return True

    def change_protection(
        self,
        image: Image | BoundImage,
        delete: bool | None = None,
    ) -> BoundAction:
        """Changes the protection configuration of the image. Can only be used on snapshots.

        :param image: :class:`BoundImage <hcloud.images.client.BoundImage>` or :class:`Image <hcloud.images.domain.Image>`
        :param delete: bool
               If true, prevents the snapshot from being deleted
        :return: :class:`BoundAction <hcloud.actions.client.BoundAction>`
        """
        data: dict[str, Any] = {}
        if delete is not None:
            data.update({"delete": delete})

        response = self._client.request(
            url=f"/images/{image.id}/actions/change_protection",
            method="POST",
            json=data,
        )
        return BoundAction(self._client.actions, response["action"])



from __future__ import annotations

from typing import Any, NamedTuple
from unittest import mock

import pytest

from hcloud.actions import ActionsPageResult
from hcloud.core import BaseDomain, BoundModelBase, ClientEntityBase, Meta


class TestBoundModelBase:
    @pytest.fixture()
    def bound_model_class(self):
        class Model(BaseDomain):
            __slots__ = ("id", "name", "description")

            def __init__(self, id, name="", description=""):
                self.id = id
                self.name = name
                self.description = description

        class BoundModel(BoundModelBase, Model):
            model = Model

        return BoundModel

    @pytest.fixture()
    def client(self):
        client = mock.MagicMock()
        return client

    def test_get_exists_model_attribute_complete_model(self, bound_model_class, client):
        bound_model = bound_model_class(
            client=client,
            data={"id": 1, "name": "name", "description": "my_description"},
        )
        description = bound_model.description
        client.get_by_id.assert_not_called()
        assert description == "my_description"

    def test_get_non_exists_model_attribute_complete_model(
        self, bound_model_class, client
    ):
        bound_model = bound_model_class(
            client=client, data={"id": 1, "name": "name", "description": "description"}
        )
        with pytest.raises(AttributeError):
            bound_model.content
        client.get_by_id.assert_not_called()

    def test_get_exists_model_attribute_incomplete_model(
        self, bound_model_class, client
    ):
        bound_model = bound_model_class(client=client, data={"id": 101}, complete=False)
        client.get_by_id.return_value = bound_model_class(
            client=client,
            data={"id": 101, "name": "name", "description": "super_description"},
        )
        description = bound_model.description
        client.get_by_id.assert_called_once_with(101)
        assert description == "super_description"
        assert bound_model.complete is True

    def test_get_filled_model_attribute_incomplete_model(
        self, bound_model_class, client
    ):
        bound_model = bound_model_class(client=client, data={"id": 101}, complete=False)
        id = bound_model.id
        client.get_by_id.assert_not_called()
        assert id == 101
        assert bound_model.complete is False

    def test_get_non_exists_model_attribute_incomplete_model(
        self, bound_model_class, client
    ):
        bound_model = bound_model_class(client=client, data={"id": 1}, complete=False)
        with pytest.raises(AttributeError):
            bound_model.content
        client.get_by_id.assert_not_called()
        assert bound_model.complete is False


class TestClientEntityBase:
    @pytest.fixture()
    def client_class_constructor(self):
        def constructor(json_content_function):
            class CandiesPageResult(NamedTuple):
                candies: list[Any]
                meta: Meta

            class CandiesClient(ClientEntityBase):
                def get_list(self, status=None, page=None, per_page=None):
                    json_content = json_content_function(page)
                    results = [
                        (r, page, status, per_page) for r in json_content["candies"]
                    ]
                    return CandiesPageResult(results, Meta.parse_meta(json_content))

            return CandiesClient(mock.MagicMock())

        return constructor

    @pytest.fixture()
    def client_class_with_actions_constructor(self):
        def constructor(json_content_function):
            class CandiesClient(ClientEntityBase):
                def get_actions_list(self, status, page=None, per_page=None):
                    json_content = json_content_function(page)
                    results = [
                        (r, page, status, per_page) for r in json_content["actions"]
                    ]
                    return ActionsPageResult(results, Meta.parse_meta(json_content))

            return CandiesClient(mock.MagicMock())

        return constructor

    def test_iter_pages_no_meta(self, client_class_constructor):
        json_content = {"candies": [1, 2]}

        def json_content_function(p):
            return json_content

        candies_client = client_class_constructor(json_content_function)

        result = candies_client._iter_pages(candies_client.get_list, status="sweet")

        assert result == [(1, 1, "sweet", 50), (2, 1, "sweet", 50)]

    def test_iter_pages_no_next_page(self, client_class_constructor):
        json_content = {
            "candies": [1, 2],
            "meta": {"pagination": {"page": 1, "per_page": 11, "next_page": None}},
        }

        def json_content_function(p):
            return json_content

        candies_client = client_class_constructor(json_content_function)

        result = candies_client._iter_pages(candies_client.get_list, status="sweet")

        assert result == [(1, 1, "sweet", 50), (2, 1, "sweet", 50)]

    def test_iter_pages_ok(self, client_class_constructor):
        def json_content_function(p):
            return {
                "candies": [10 + p, 20 + p],
                "meta": {
                    "pagination": {
                        "page": p,
                        "per_page": 11,
                        "next_page": p + 1 if p < 3 else None,
                    }
                },
            }

        candies_client = client_class_constructor(json_content_function)

        result = candies_client._iter_pages(candies_client.get_list, status="sweet")

        assert result == [
            (11, 1, "sweet", 50),
            (21, 1, "sweet", 50),
            (12, 2, "sweet", 50),
            (22, 2, "sweet", 50),
            (13, 3, "sweet", 50),
            (23, 3, "sweet", 50),
        ]

    def test_get_actions_ok(self, client_class_with_actions_constructor):
        def json_content_function(p):
            return {
                "actions": [10 + p, 20 + p],
                "meta": {
                    "pagination": {
                        "page": p,
                        "per_page": 11,
                        "next_page": p + 1 if p < 3 else None,
                    }
                },
            }

        candies_client = client_class_with_actions_constructor(json_content_function)

        result = candies_client._iter_pages(
            candies_client.get_actions_list, status="sweet"
        )

        assert result == [
            (11, 1, "sweet", 50),
            (21, 1, "sweet", 50),
            (12, 2, "sweet", 50),
            (22, 2, "sweet", 50),
            (13, 3, "sweet", 50),
            (23, 3, "sweet", 50),
        ]

    def test_get_first_by_result_exists(self, client_class_constructor):
        json_content = {"candies": [1]}

        def json_content_function(p):
            return json_content

        candies_client = client_class_constructor(json_content_function)

        result = candies_client._get_first_by(status="sweet")

        assert result == (1, None, "sweet", None)

    def test_get_first_by_result_does_not_exist(self, client_class_constructor):
        json_content = {"candies": []}

        def json_content_function(p):
            return json_content

        candies_client = client_class_constructor(json_content_function)

        result = candies_client._get_first_by(status="sweet")

        assert result is None



from __future__ import annotations

from unittest import mock

import pytest

from hcloud.actions import BoundAction
from hcloud.datacenters import BoundDatacenter, Datacenter
from hcloud.primary_ips import BoundPrimaryIP, PrimaryIP, PrimaryIPsClient


class TestBoundPrimaryIP:
    @pytest.fixture()
    def bound_primary_ip(self, hetzner_client):
        return BoundPrimaryIP(client=hetzner_client.primary_ips, data=dict(id=14))

    def test_bound_primary_ip_init(self, primary_ip_response):
        bound_primary_ip = BoundPrimaryIP(
            client=mock.MagicMock(), data=primary_ip_response["primary_ip"]
        )

        assert bound_primary_ip.id == 42
        assert bound_primary_ip.name == "my-resource"
        assert bound_primary_ip.ip == "131.232.99.1"
        assert bound_primary_ip.type == "ipv4"
        assert bound_primary_ip.protection == {"delete": False}
        assert bound_primary_ip.labels == {}
        assert bound_primary_ip.blocked is False

        assert bound_primary_ip.assignee_id == 17
        assert bound_primary_ip.assignee_type == "server"

        assert isinstance(bound_primary_ip.datacenter, BoundDatacenter)
        assert bound_primary_ip.datacenter.id == 42
        assert bound_primary_ip.datacenter.name == "fsn1-dc8"
        assert bound_primary_ip.datacenter.description == "Falkenstein DC Park 8"
        assert bound_primary_ip.datacenter.location.country == "DE"
        assert bound_primary_ip.datacenter.location.city == "Falkenstein"
        assert bound_primary_ip.datacenter.location.latitude == 50.47612
        assert bound_primary_ip.datacenter.location.longitude == 12.370071

    def test_update(self, hetzner_client, bound_primary_ip, response_update_primary_ip):
        hetzner_client.request.return_value = response_update_primary_ip
        primary_ip = bound_primary_ip.update(auto_delete=True, name="my-resource")
        hetzner_client.request.assert_called_with(
            url="/primary_ips/14",
            method="PUT",
            json={"auto_delete": True, "name": "my-resource"},
        )

        assert primary_ip.id == 42
        assert primary_ip.auto_delete is True

    def test_delete(self, hetzner_client, bound_primary_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        delete_success = bound_primary_ip.delete()
        hetzner_client.request.assert_called_with(
            url="/primary_ips/14", method="DELETE"
        )

        assert delete_success is True

    def test_change_protection(self, hetzner_client, bound_primary_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_primary_ip.change_protection(True)
        hetzner_client.request.assert_called_with(
            url="/primary_ips/14/actions/change_protection",
            method="POST",
            json={"delete": True},
        )

        assert action.id == 1
        assert action.progress == 0

    def test_assign(self, hetzner_client, bound_primary_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_primary_ip.assign(assignee_id=12, assignee_type="server")
        hetzner_client.request.assert_called_with(
            url="/primary_ips/14/actions/assign",
            method="POST",
            json={"assignee_id": 12, "assignee_type": "server"},
        )
        assert action.id == 1
        assert action.progress == 0

    def test_unassign(self, hetzner_client, bound_primary_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_primary_ip.unassign()
        hetzner_client.request.assert_called_with(
            url="/primary_ips/14/actions/unassign", method="POST"
        )
        assert action.id == 1
        assert action.progress == 0

    def test_change_dns_ptr(self, hetzner_client, bound_primary_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_primary_ip.change_dns_ptr("1.2.3.4", "server02.example.com")
        hetzner_client.request.assert_called_with(
            url="/primary_ips/14/actions/change_dns_ptr",
            method="POST",
            json={"ip": "1.2.3.4", "dns_ptr": "server02.example.com"},
        )
        assert action.id == 1
        assert action.progress == 0


class TestPrimaryIPsClient:
    @pytest.fixture()
    def primary_ips_client(self):
        return PrimaryIPsClient(client=mock.MagicMock())

    def test_get_by_id(self, primary_ips_client, primary_ip_response):
        primary_ips_client._client.request.return_value = primary_ip_response
        bound_primary_ip = primary_ips_client.get_by_id(1)
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/1", method="GET"
        )
        assert bound_primary_ip._client is primary_ips_client
        assert bound_primary_ip.id == 42

    def test_get_by_name(self, primary_ips_client, one_primary_ips_response):
        primary_ips_client._client.request.return_value = one_primary_ips_response
        bound_primary_ip = primary_ips_client.get_by_name("my-resource")
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips", method="GET", params={"name": "my-resource"}
        )
        assert bound_primary_ip._client is primary_ips_client
        assert bound_primary_ip.id == 42
        assert bound_primary_ip.name == "my-resource"

    @pytest.mark.parametrize("params", [{"label_selector": "label1"}])
    def test_get_all(self, primary_ips_client, all_primary_ips_response, params):
        primary_ips_client._client.request.return_value = all_primary_ips_response
        bound_primary_ips = primary_ips_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips", method="GET", params=params
        )

        assert len(bound_primary_ips) == 1

        bound_primary_ip1 = bound_primary_ips[0]

        assert bound_primary_ip1._client is primary_ips_client
        assert bound_primary_ip1.id == 42
        assert bound_primary_ip1.name == "my-resource"

    def test_create_with_datacenter(self, primary_ips_client, primary_ip_response):
        primary_ips_client._client.request.return_value = primary_ip_response
        response = primary_ips_client.create(
            type="ipv6", name="my-resource", datacenter=Datacenter(name="datacenter")
        )
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips",
            method="POST",
            json={
                "name": "my-resource",
                "type": "ipv6",
                "datacenter": "datacenter",
                "auto_delete": False,
                "assignee_type": "server",
            },
        )

        bound_primary_ip = response.primary_ip
        action = response.action

        assert bound_primary_ip._client is primary_ips_client
        assert bound_primary_ip.id == 42
        assert bound_primary_ip.name == "my-resource"
        assert action is None

    def test_create_with_assignee_id(
        self, primary_ips_client, primary_ip_create_response
    ):
        primary_ips_client._client.request.return_value = primary_ip_create_response
        response = primary_ips_client.create(
            type="ipv6",
            name="my-ip",
            assignee_id=1,
            assignee_type="server",
            datacenter=Datacenter(name="datacenter"),
        )
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips",
            method="POST",
            json={
                "type": "ipv6",
                "assignee_id": 1,
                "assignee_type": "server",
                "name": "my-ip",
                "auto_delete": False,
                "datacenter": "datacenter",
            },
        )
        bound_primary_ip = response.primary_ip
        action = response.action

        assert bound_primary_ip._client is primary_ips_client
        assert bound_primary_ip.id == 42
        assert bound_primary_ip.name == "my-ip"
        assert bound_primary_ip.assignee_id == 17
        assert action.id == 13

    @pytest.mark.parametrize(
        "primary_ip", [PrimaryIP(id=1), BoundPrimaryIP(mock.MagicMock(), dict(id=1))]
    )
    def test_update(self, primary_ips_client, primary_ip, response_update_primary_ip):
        primary_ips_client._client.request.return_value = response_update_primary_ip
        primary_ip = primary_ips_client.update(
            primary_ip, auto_delete=True, name="my-resource"
        )
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/1",
            method="PUT",
            json={"auto_delete": True, "name": "my-resource"},
        )

        assert primary_ip.id == 42
        assert primary_ip.auto_delete is True
        assert primary_ip.name == "my-resource"

    @pytest.mark.parametrize(
        "primary_ip", [PrimaryIP(id=1), BoundPrimaryIP(mock.MagicMock(), dict(id=1))]
    )
    def test_change_protection(self, primary_ips_client, primary_ip, generic_action):
        primary_ips_client._client.request.return_value = generic_action
        action = primary_ips_client.change_protection(primary_ip, True)
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/1/actions/change_protection",
            method="POST",
            json={"delete": True},
        )

        assert action.id == 1
        assert action.progress == 0

    @pytest.mark.parametrize(
        "primary_ip", [PrimaryIP(id=1), BoundPrimaryIP(mock.MagicMock(), dict(id=1))]
    )
    def test_delete(self, primary_ips_client, primary_ip, generic_action):
        primary_ips_client._client.request.return_value = generic_action
        delete_success = primary_ips_client.delete(primary_ip)
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/1", method="DELETE"
        )

        assert delete_success is True

    @pytest.mark.parametrize(
        "assignee_id,assignee_type,primary_ip",
        [
            (1, "server", PrimaryIP(id=12)),
            (1, "server", BoundPrimaryIP(mock.MagicMock(), dict(id=12))),
        ],
    )
    def test_assign(
        self, primary_ips_client, assignee_id, assignee_type, primary_ip, generic_action
    ):
        primary_ips_client._client.request.return_value = generic_action
        action = primary_ips_client.assign(primary_ip, assignee_id, assignee_type)
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/12/actions/assign",
            method="POST",
            json={"assignee_id": 1, "assignee_type": "server"},
        )
        assert action.id == 1
        assert action.progress == 0

    @pytest.mark.parametrize(
        "primary_ip", [PrimaryIP(id=12), BoundPrimaryIP(mock.MagicMock(), dict(id=12))]
    )
    def test_unassign(self, primary_ips_client, primary_ip, generic_action):
        primary_ips_client._client.request.return_value = generic_action
        action = primary_ips_client.unassign(primary_ip)
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/12/actions/unassign", method="POST"
        )
        assert action.id == 1
        assert action.progress == 0

    @pytest.mark.parametrize(
        "primary_ip", [PrimaryIP(id=12), BoundPrimaryIP(mock.MagicMock(), dict(id=12))]
    )
    def test_change_dns_ptr(self, primary_ips_client, primary_ip, generic_action):
        primary_ips_client._client.request.return_value = generic_action
        action = primary_ips_client.change_dns_ptr(
            primary_ip, "1.2.3.4", "server02.example.com"
        )
        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/12/actions/change_dns_ptr",
            method="POST",
            json={"ip": "1.2.3.4", "dns_ptr": "server02.example.com"},
        )
        assert action.id == 1
        assert action.progress == 0

    def test_actions_get_by_id(self, primary_ips_client, response_get_actions):
        primary_ips_client._client.request.return_value = {
            "action": response_get_actions["actions"][0]
        }
        action = primary_ips_client.actions.get_by_id(13)

        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/actions/13", method="GET"
        )

        assert isinstance(action, BoundAction)
        assert action._client == primary_ips_client._client.actions
        assert action.id == 13
        assert action.command == "assign_primary_ip"

    def test_actions_get_list(self, primary_ips_client, response_get_actions):
        primary_ips_client._client.request.return_value = response_get_actions
        result = primary_ips_client.actions.get_list()

        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/actions",
            method="GET",
            params={},
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == primary_ips_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "assign_primary_ip"

    def test_actions_get_all(self, primary_ips_client, response_get_actions):
        primary_ips_client._client.request.return_value = response_get_actions
        actions = primary_ips_client.actions.get_all()

        primary_ips_client._client.request.assert_called_with(
            url="/primary_ips/actions",
            method="GET",
            params={"page": 1, "per_page": 50},
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == primary_ips_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "assign_primary_ip"



from __future__ import annotations

import os
import sys
from datetime import datetime

sys.path.insert(0, os.path.abspath(".."))
from hcloud.__version__ import VERSION  # noqa

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Hetzner Cloud Python"
author = "Hetzner Cloud GmbH"
copyright = f"{datetime.now().year}, {author}"

version = VERSION
release = VERSION

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser", "sphinx.ext.autodoc", "sphinx.ext.viewcode"]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# A boolean that decides whether module names are prepended to all object names (for
# object types where a “module” of some kind is defined), e.g. for py:function
# directives. Default is True.
add_module_names = False

# Myst Parser
myst_enable_extensions = ["colon_fence"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

html_logo = "_static/logo-hetzner.svg"
html_favicon = "_static/favicon.png"
# Theme options are theme-specific and customize the look and feel of a theme further.
# For a list of options available for each theme, see the documentation.
html_theme_options = {
    "logo_only": True,
    "style_nav_header_background": "#fff",
}
html_css_files = [
    "custom.css",
]



from __future__ import annotations

from .client import (  # noqa: F401
    BoundLoadBalancerType,
    LoadBalancerTypesClient,
    LoadBalancerTypesPageResult,
)
from .domain import LoadBalancerType  # noqa: F401



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from dateutil.parser import isoparse

from ..core import BaseDomain

if TYPE_CHECKING:
    from ..actions import BoundAction
    from ..datacenters import BoundDatacenter
    from ..firewalls import BoundFirewall
    from ..floating_ips import BoundFloatingIP
    from ..images import BoundImage
    from ..isos import BoundIso
    from ..metrics import Metrics
    from ..networks import BoundNetwork
    from ..placement_groups import BoundPlacementGroup
    from ..primary_ips import BoundPrimaryIP, PrimaryIP
    from ..server_types import BoundServerType
    from ..volumes import BoundVolume
    from .client import BoundServer


class Server(BaseDomain):
    """Server Domain

    :param id: int
           ID of the server
    :param name: str
           Name of the server (must be unique per project and a valid hostname as per RFC 1123)
    :param status: str
           Status of the server Choices: `running`, `initializing`, `starting`, `stopping`, `off`, `deleting`, `migrating`, `rebuilding`, `unknown`
    :param created: datetime
           Point in time when the server was created
    :param public_net: :class:`PublicNetwork <hcloud.servers.domain.PublicNetwork>`
           Public network information.
    :param server_type: :class:`BoundServerType <hcloud.server_types.client.BoundServerType>`
    :param datacenter: :class:`BoundDatacenter <hcloud.datacenters.client.BoundDatacenter>`
    :param image: :class:`BoundImage <hcloud.images.client.BoundImage>`, None
    :param iso: :class:`BoundIso <hcloud.isos.client.BoundIso>`, None
    :param rescue_enabled: bool
           True if rescue mode is enabled: Server will then boot into rescue system on next reboot.
    :param locked: bool
           True if server has been locked and is not available to user.
    :param backup_window: str, None
           Time window (UTC) in which the backup will run, or None if the backups are not enabled
    :param outgoing_traffic: int, None
           Outbound Traffic for the current billing period in bytes
    :param ingoing_traffic: int, None
           Inbound Traffic for the current billing period in bytes
    :param included_traffic: int
           Free Traffic for the current billing period in bytes
    :param primary_disk_size: int
           Size of the primary Disk
    :param protection: dict
           Protection configuration for the server
    :param labels: dict
            User-defined labels (key-value pairs)
    :param volumes: List[:class:`BoundVolume <hcloud.volumes.client.BoundVolume>`]
            Volumes assigned to this server.
    :param private_net: List[:class:`PrivateNet <hcloud.servers.domain.PrivateNet>`]
            Private networks information.
    """

    STATUS_RUNNING = "running"
    """Server Status running"""
    STATUS_INIT = "initializing"
    """Server Status initializing"""
    STATUS_STARTING = "starting"
    """Server Status starting"""
    STATUS_STOPPING = "stopping"
    """Server Status stopping"""
    STATUS_OFF = "off"
    """Server Status off"""
    STATUS_DELETING = "deleting"
    """Server Status deleting"""
    STATUS_MIGRATING = "migrating"
    """Server Status migrating"""
    STATUS_REBUILDING = "rebuilding"
    """Server Status rebuilding"""
    STATUS_UNKNOWN = "unknown"
    """Server Status unknown"""
    __slots__ = (
        "id",
        "name",
        "status",
        "public_net",
        "server_type",
        "datacenter",
        "image",
        "iso",
        "rescue_enabled",
        "locked",
        "backup_window",
        "outgoing_traffic",
        "ingoing_traffic",
        "included_traffic",
        "protection",
        "labels",
        "volumes",
        "private_net",
        "created",
        "primary_disk_size",
        "placement_group",
    )

    # pylint: disable=too-many-locals
    def __init__(
        self,
        id: int,
        name: str | None = None,
        status: str | None = None,
        created: str | None = None,
        public_net: PublicNetwork | None = None,
        server_type: BoundServerType | None = None,
        datacenter: BoundDatacenter | None = None,
        image: BoundImage | None = None,
        iso: BoundIso | None = None,
        rescue_enabled: bool | None = None,
        locked: bool | None = None,
        backup_window: str | None = None,
        outgoing_traffic: int | None = None,
        ingoing_traffic: int | None = None,
        included_traffic: int | None = None,
        protection: dict | None = None,
        labels: dict[str, str] | None = None,
        volumes: list[BoundVolume] | None = None,
        private_net: list[PrivateNet] | None = None,
        primary_disk_size: int | None = None,
        placement_group: BoundPlacementGroup | None = None,
    ):
        self.id = id
        self.name = name
        self.status = status
        self.created = isoparse(created) if created else None
        self.public_net = public_net
        self.server_type = server_type
        self.datacenter = datacenter
        self.image = image
        self.iso = iso
        self.rescue_enabled = rescue_enabled
        self.locked = locked
        self.backup_window = backup_window
        self.outgoing_traffic = outgoing_traffic
        self.ingoing_traffic = ingoing_traffic
        self.included_traffic = included_traffic
        self.protection = protection
        self.labels = labels
        self.volumes = volumes
        self.private_net = private_net
        self.primary_disk_size = primary_disk_size
        self.placement_group = placement_group


class CreateServerResponse(BaseDomain):
    """Create Server Response Domain

    :param server: :class:`BoundServer <hcloud.servers.client.BoundServer>`
           The created server
    :param action: :class:`BoundAction <hcloud.actions.client.BoundAction>`
           Shows the progress of the server creation
    :param next_actions: List[:class:`BoundAction <hcloud.actions.client.BoundAction>`]
           Additional actions like a `start_server` action after the server creation
    :param root_password: str, None
           The root password of the server if no SSH-Key was given on server creation
    """

    __slots__ = ("server", "action", "next_actions", "root_password")

    def __init__(
        self,
        server: BoundServer,
        action: BoundAction,
        next_actions: list[BoundAction],
        root_password: str | None,
    ):
        self.server = server
        self.action = action
        self.next_actions = next_actions
        self.root_password = root_password


class ResetPasswordResponse(BaseDomain):
    """Reset Password Response Domain

    :param action: :class:`BoundAction <hcloud.actions.client.BoundAction>`
           Shows the progress of the server passwort reset action
    :param root_password: str
           The root password of the server
    """

    __slots__ = ("action", "root_password")

    def __init__(
        self,
        action: BoundAction,
        root_password: str,
    ):
        self.action = action
        self.root_password = root_password


class EnableRescueResponse(BaseDomain):
    """Enable Rescue Response Domain

    :param action: :class:`BoundAction <hcloud.actions.client.BoundAction>`
           Shows the progress of the server enable rescue action
    :param root_password: str
           The root password of the server in the rescue mode
    """

    __slots__ = ("action", "root_password")

    def __init__(
        self,
        action: BoundAction,
        root_password: str,
    ):
        self.action = action
        self.root_password = root_password


class RequestConsoleResponse(BaseDomain):
    """Request Console Response Domain

    :param action: :class:`BoundAction <hcloud.actions.client.BoundAction>`
           Shows the progress of the server request console action
    :param wss_url: str
           URL of websocket proxy to use. This includes a token which is valid for a limited time only.
    :param password: str
           VNC password to use for this connection. This password only works in combination with a wss_url with valid token.
    """

    __slots__ = ("action", "wss_url", "password")

    def __init__(
        self,
        action: BoundAction,
        wss_url: str,
        password: str,
    ):
        self.action = action
        self.wss_url = wss_url
        self.password = password


class RebuildResponse(BaseDomain):
    """Rebuild Response Domain

    :param action: Shows the progress of the server rebuild action
    :param root_password: The root password of the server when not using SSH keys
    """

    __slots__ = ("action", "root_password")

    def __init__(
        self,
        action: BoundAction,
        root_password: str | None,
    ):
        self.action = action
        self.root_password = root_password


class PublicNetwork(BaseDomain):
    """Public Network Domain

    :param ipv4: :class:`IPv4Address <hcloud.servers.domain.IPv4Address>`
    :param ipv6: :class:`IPv6Network <hcloud.servers.domain.IPv6Network>`
    :param floating_ips: List[:class:`BoundFloatingIP <hcloud.floating_ips.client.BoundFloatingIP>`]
    :param primary_ipv4: :class:`BoundPrimaryIP <hcloud.primary_ips.domain.BoundPrimaryIP>`
    :param primary_ipv6: :class:`BoundPrimaryIP <hcloud.primary_ips.domain.BoundPrimaryIP>`
    :param firewalls: List[:class:`PublicNetworkFirewall <hcloud.servers.client.PublicNetworkFirewall>`]
    """

    __slots__ = (
        "ipv4",
        "ipv6",
        "floating_ips",
        "firewalls",
        "primary_ipv4",
        "primary_ipv6",
    )

    def __init__(
        self,
        ipv4: IPv4Address,
        ipv6: IPv6Network,
        floating_ips: list[BoundFloatingIP],
        primary_ipv4: BoundPrimaryIP | None,
        primary_ipv6: BoundPrimaryIP | None,
        firewalls: list[PublicNetworkFirewall] | None = None,
    ):
        self.ipv4 = ipv4
        self.ipv6 = ipv6
        self.floating_ips = floating_ips
        self.firewalls = firewalls
        self.primary_ipv4 = primary_ipv4
        self.primary_ipv6 = primary_ipv6


class PublicNetworkFirewall(BaseDomain):
    """Public Network Domain

    :param firewall: :class:`BoundFirewall <hcloud.firewalls.domain.BoundFirewall>`
    :param status: str
    """

    __slots__ = ("firewall", "status")

    STATUS_APPLIED = "applied"
    """Public Network Firewall Status applied"""
    STATUS_PENDING = "pending"
    """Public Network Firewall Status pending"""

    def __init__(
        self,
        firewall: BoundFirewall,
        status: str,
    ):
        self.firewall = firewall
        self.status = status


class IPv4Address(BaseDomain):
    """IPv4 Address Domain

    :param ip: str
           The IPv4 Address
    :param blocked: bool
           Determine if the IP is blocked
    :param dns_ptr: str
           DNS PTR for the ip
    """

    __slots__ = ("ip", "blocked", "dns_ptr")

    def __init__(
        self,
        ip: str,
        blocked: bool,
        dns_ptr: str,
    ):
        self.ip = ip
        self.blocked = blocked
        self.dns_ptr = dns_ptr


class IPv6Network(BaseDomain):
    """IPv6 Network Domain

    :param ip: str
           The IPv6 Network as CIDR Notation
    :param blocked: bool
           Determine if the Network is blocked
    :param dns_ptr: dict
           DNS PTR Records for the Network as Dict
    :param network: str
           The network without the network mask
    :param network_mask: str
           The network mask
    """

    __slots__ = ("ip", "blocked", "dns_ptr", "network", "network_mask")

    def __init__(
        self,
        ip: str,
        blocked: bool,
        dns_ptr: list,
    ):
        self.ip = ip
        self.blocked = blocked
        self.dns_ptr = dns_ptr
        ip_parts = self.ip.split("/")  # 2001:db8::/64 to 2001:db8:: and 64
        self.network = ip_parts[0]
        self.network_mask = ip_parts[1]


class PrivateNet(BaseDomain):
    """PrivateNet Domain

    :param network: :class:`BoundNetwork <hcloud.networks.client.BoundNetwork>`
           The network the server is attached to
    :param ip: str
           The main IP Address of the server in the Network
    :param alias_ips: List[str]
           The alias ips for a server
    :param mac_address: str
           The mac address of the interface on the server
    """

    __slots__ = ("network", "ip", "alias_ips", "mac_address")

    def __init__(
        self,
        network: BoundNetwork,
        ip: str,
        alias_ips: list[str],
        mac_address: str,
    ):
        self.network = network
        self.ip = ip
        self.alias_ips = alias_ips
        self.mac_address = mac_address


class ServerCreatePublicNetwork(BaseDomain):
    """Server Create Public Network Domain

    :param ipv4: Optional[:class:`PrimaryIP <hcloud.primary_ips.domain.PrimaryIP>`]
    :param ipv6: Optional[:class:`PrimaryIP <hcloud.primary_ips.domain.PrimaryIP>`]
    :param enable_ipv4: bool
    :param enable_ipv6: bool
    """

    __slots__ = ("ipv4", "ipv6", "enable_ipv4", "enable_ipv6")

    def __init__(
        self,
        ipv4: PrimaryIP | None = None,
        ipv6: PrimaryIP | None = None,
        enable_ipv4: bool = True,
        enable_ipv6: bool = True,
    ):
        self.ipv4 = ipv4
        self.ipv6 = ipv6
        self.enable_ipv4 = enable_ipv4
        self.enable_ipv6 = enable_ipv6


MetricsType = Literal[
    "cpu",
    "disk",
    "network",
]


class GetMetricsResponse(BaseDomain):
    """Get a Server Metrics Response Domain

    :param metrics: The Server metrics
    """

    __slots__ = ("metrics",)

    def __init__(
        self,
        metrics: Metrics,
    ):
        self.metrics = metrics



from __future__ import annotations

from .client import (  # noqa: F401
    BoundFloatingIP,
    FloatingIPsClient,
    FloatingIPsPageResult,
)
from .domain import CreateFloatingIPResponse, FloatingIP  # noqa: F401



from __future__ import annotations

from unittest import mock  # noqa: F401

import pytest  # noqa: F401

from hcloud.locations import LocationsClient


class TestLocationsClient:
    @pytest.fixture()
    def locations_client(self):
        return LocationsClient(client=mock.MagicMock())

    def test_get_by_id(self, locations_client, location_response):
        locations_client._client.request.return_value = location_response
        location = locations_client.get_by_id(1)
        locations_client._client.request.assert_called_with(
            url="/locations/1", method="GET"
        )
        assert location._client is locations_client
        assert location.id == 1
        assert location.name == "fsn1"
        assert location.network_zone == "eu-central"

    @pytest.mark.parametrize(
        "params", [{"name": "fsn1", "page": 1, "per_page": 10}, {"name": ""}, {}]
    )
    def test_get_list(self, locations_client, two_locations_response, params):
        locations_client._client.request.return_value = two_locations_response
        result = locations_client.get_list(**params)
        locations_client._client.request.assert_called_with(
            url="/locations", method="GET", params=params
        )

        locations = result.locations
        assert result.meta is None

        assert len(locations) == 2

        location1 = locations[0]
        location2 = locations[1]

        assert location1._client is locations_client
        assert location1.id == 1
        assert location1.name == "fsn1"
        assert location1.network_zone == "eu-central"

        assert location2._client is locations_client
        assert location2.id == 2
        assert location2.name == "nbg1"
        assert location2.network_zone == "eu-central"

    @pytest.mark.parametrize("params", [{"name": "fsn1"}, {}])
    def test_get_all(self, locations_client, two_locations_response, params):
        locations_client._client.request.return_value = two_locations_response
        locations = locations_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        locations_client._client.request.assert_called_with(
            url="/locations", method="GET", params=params
        )

        assert len(locations) == 2

        location1 = locations[0]
        location2 = locations[1]

        assert location1._client is locations_client
        assert location1.id == 1
        assert location1.name == "fsn1"
        assert location1.network_zone == "eu-central"

        assert location2._client is locations_client
        assert location2.id == 2
        assert location2.name == "nbg1"
        assert location2.network_zone == "eu-central"

    def test_get_by_name(self, locations_client, one_locations_response):
        locations_client._client.request.return_value = one_locations_response
        location = locations_client.get_by_name("fsn1")

        params = {"name": "fsn1"}

        locations_client._client.request.assert_called_with(
            url="/locations", method="GET", params=params
        )

        assert location._client is locations_client
        assert location.id == 1
        assert location.name == "fsn1"
        assert location.network_zone == "eu-central"



from __future__ import annotations

from unittest import mock

import pytest

from hcloud.actions import (
    Action,
    ActionFailedException,
    ActionsClient,
    ActionTimeoutException,
    BoundAction,
    ResourceActionsClient,
)


class TestBoundAction:
    @pytest.fixture()
    def bound_running_action(self, mocked_requests):
        action_client = ActionsClient(client=mocked_requests)
        # Speed up tests that run `wait_until_finished`
        action_client._client.poll_interval = 0.1

        return BoundAction(
            client=action_client,
            data=dict(id=14, status=Action.STATUS_RUNNING),
        )

    def test_wait_until_finished(
        self, bound_running_action, mocked_requests, running_action, successfully_action
    ):
        mocked_requests.request.side_effect = [running_action, successfully_action]
        bound_running_action.wait_until_finished()
        mocked_requests.request.assert_called_with(url="/actions/2", method="GET")

        assert bound_running_action.status == "success"
        assert mocked_requests.request.call_count == 2

    def test_wait_until_finished_with_error(
        self, bound_running_action, mocked_requests, running_action, failed_action
    ):
        mocked_requests.request.side_effect = [running_action, failed_action]
        with pytest.raises(ActionFailedException) as exception_info:
            bound_running_action.wait_until_finished()

        assert bound_running_action.status == "error"
        assert exception_info.value.action.id == 2

    def test_wait_until_finished_max_retries(
        self, bound_running_action, mocked_requests, running_action, successfully_action
    ):
        mocked_requests.request.side_effect = [
            running_action,
            running_action,
            successfully_action,
        ]
        with pytest.raises(ActionTimeoutException) as exception_info:
            bound_running_action.wait_until_finished(max_retries=1)

        assert bound_running_action.status == "running"
        assert exception_info.value.action.id == 2
        assert mocked_requests.request.call_count == 1


class TestResourceActionsClient:
    @pytest.fixture()
    def actions_client(self):
        return ResourceActionsClient(client=mock.MagicMock(), resource="/resource")

    def test_get_by_id(self, actions_client, generic_action):
        actions_client._client.request.return_value = generic_action
        action = actions_client.get_by_id(1)
        actions_client._client.request.assert_called_with(
            url="/resource/actions/1", method="GET"
        )
        assert action._client == actions_client._client.actions
        assert action.id == 1
        assert action.command == "stop_server"

    @pytest.mark.parametrize(
        "params",
        [{}, {"status": ["active"], "sort": ["status"], "page": 2, "per_page": 10}],
    )
    def test_get_list(self, actions_client, generic_action_list, params):
        actions_client._client.request.return_value = generic_action_list
        result = actions_client.get_list(**params)
        actions_client._client.request.assert_called_with(
            url="/resource/actions", method="GET", params=params
        )

        assert result.meta is None

        actions = result.actions
        assert len(actions) == 2

        action1 = actions[0]
        action2 = actions[1]

        assert action1._client == actions_client._client.actions
        assert action1.id == 1
        assert action1.command == "start_server"

        assert action2._client == actions_client._client.actions
        assert action2.id == 2
        assert action2.command == "stop_server"

    @pytest.mark.parametrize("params", [{}, {"status": ["active"], "sort": ["status"]}])
    def test_get_all(self, actions_client, generic_action_list, params):
        actions_client._client.request.return_value = generic_action_list
        actions = actions_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        actions_client._client.request.assert_called_with(
            url="/resource/actions", method="GET", params=params
        )

        assert len(actions) == 2

        action1 = actions[0]
        action2 = actions[1]

        assert action1._client == actions_client._client.actions
        assert action1.id == 1
        assert action1.command == "start_server"

        assert action2._client == actions_client._client.actions
        assert action2.id == 2
        assert action2.command == "stop_server"


class TestActionsClient:
    @pytest.fixture()
    def actions_client(self):
        return ActionsClient(client=mock.MagicMock())

    def test_get_by_id(self, actions_client, generic_action):
        actions_client._client.request.return_value = generic_action
        action = actions_client.get_by_id(1)
        actions_client._client.request.assert_called_with(
            url="/actions/1", method="GET"
        )
        assert action._client == actions_client._client.actions
        assert action.id == 1
        assert action.command == "stop_server"

    @pytest.mark.parametrize(
        "params",
        [{}, {"status": ["active"], "sort": ["status"], "page": 2, "per_page": 10}],
    )
    def test_get_list(self, actions_client, generic_action_list, params):
        actions_client._client.request.return_value = generic_action_list
        with pytest.deprecated_call():
            result = actions_client.get_list(**params)
        actions_client._client.request.assert_called_with(
            url="/actions", method="GET", params=params
        )

        assert result.meta is None

        actions = result.actions
        assert len(actions) == 2

        action1 = actions[0]
        action2 = actions[1]

        assert action1._client == actions_client._client.actions
        assert action1.id == 1
        assert action1.command == "start_server"

        assert action2._client == actions_client._client.actions
        assert action2.id == 2
        assert action2.command == "stop_server"

    @pytest.mark.parametrize("params", [{}, {"status": ["active"], "sort": ["status"]}])
    def test_get_all(self, actions_client, generic_action_list, params):
        actions_client._client.request.return_value = generic_action_list
        with pytest.deprecated_call():
            actions = actions_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        actions_client._client.request.assert_called_with(
            url="/actions", method="GET", params=params
        )

        assert len(actions) == 2

        action1 = actions[0]
        action2 = actions[1]

        assert action1._client == actions_client._client.actions
        assert action1.id == 1
        assert action1.command == "start_server"

        assert action2._client == actions_client._client.actions
        assert action2.id == 2
        assert action2.command == "stop_server"



from __future__ import annotations

from .client import (  # noqa: F401
    BoundPlacementGroup,
    PlacementGroupsClient,
    PlacementGroupsPageResult,
)
from .domain import CreatePlacementGroupResponse, PlacementGroup  # noqa: F401



from __future__ import annotations

import pytest
from dateutil.parser import isoparse

from hcloud.core import BaseDomain, DomainIdentityMixin, Meta, Pagination


class TestMeta:
    @pytest.mark.parametrize("json_content", [None, "", {}])
    def test_parse_meta_empty_json(self, json_content):
        result = Meta.parse_meta(json_content)
        assert result is None

    def test_parse_meta_json_no_paginaton(self):
        json_content = {"meta": {}}
        result = Meta.parse_meta(json_content)
        assert isinstance(result, Meta)
        assert result.pagination is None

    def test_parse_meta_json_ok(self):
        json_content = {
            "meta": {
                "pagination": {
                    "page": 2,
                    "per_page": 10,
                    "previous_page": 1,
                    "next_page": 3,
                    "last_page": 10,
                    "total_entries": 100,
                }
            }
        }
        result = Meta.parse_meta(json_content)
        assert isinstance(result, Meta)
        assert isinstance(result.pagination, Pagination)
        assert result.pagination.page == 2
        assert result.pagination.per_page == 10
        assert result.pagination.next_page == 3
        assert result.pagination.last_page == 10
        assert result.pagination.total_entries == 100


class SomeDomain(BaseDomain, DomainIdentityMixin):
    __slots__ = ("id", "name")

    def __init__(self, id=None, name=None):
        self.id = id
        self.name = name


class TestDomainIdentityMixin:
    @pytest.mark.parametrize(
        "domain,expected_result",
        [
            (SomeDomain(id=1, name="name"), 1),
            (SomeDomain(id=1), 1),
            (SomeDomain(name="name"), "name"),
        ],
    )
    def test_id_or_name_ok(self, domain, expected_result):
        assert domain.id_or_name == expected_result

    def test_id_or_name_exception(self):
        domain = SomeDomain()

        with pytest.raises(ValueError) as exception_info:
            domain.id_or_name
        error = exception_info.value
        assert str(error) == "id or name must be set"


class ActionDomain(BaseDomain, DomainIdentityMixin):
    __slots__ = ("id", "name", "started")

    def __init__(self, id, name="name1", started=None):
        self.id = id
        self.name = name
        self.started = isoparse(started) if started else None


class SomeOtherDomain(BaseDomain):
    __slots__ = ("id", "name", "child")

    def __init__(self, id=None, name=None, child=None):
        self.id = id
        self.name = name
        self.child = child


class TestBaseDomain:
    @pytest.mark.parametrize(
        "data_dict,expected_result",
        [
            ({"id": 1}, {"id": 1, "name": "name1", "started": None}),
            ({"id": 2, "name": "name2"}, {"id": 2, "name": "name2", "started": None}),
            (
                {"id": 3, "foo": "boo", "description": "new"},
                {"id": 3, "name": "name1", "started": None},
            ),
            (
                {
                    "id": 4,
                    "foo": "boo",
                    "description": "new",
                    "name": "name-name3",
                    "started": "2016-01-30T23:50+00:00",
                },
                {
                    "id": 4,
                    "name": "name-name3",
                    "started": isoparse("2016-01-30T23:50+00:00"),
                },
            ),
        ],
    )
    def test_from_dict_ok(self, data_dict, expected_result):
        model = ActionDomain.from_dict(data_dict)
        for k, v in expected_result.items():
            assert getattr(model, k) == v

    @pytest.mark.parametrize(
        "data,expected",
        [
            (
                SomeOtherDomain(id=1, name="name1"),
                "SomeOtherDomain(id=1, name='name1', child=None)",
            ),
            (
                SomeOtherDomain(
                    id=2,
                    name="name2",
                    child=SomeOtherDomain(id=3, name="name3"),
                ),
                "SomeOtherDomain(id=2, name='name2', child=SomeOtherDomain(id=3, name='name3', child=None))",
            ),
        ],
    )
    def test_repr_ok(self, data, expected):
        assert data.__repr__() == expected



from __future__ import annotations

from unittest import mock

import pytest

from hcloud import Client


@pytest.fixture(autouse=True, scope="function")
def mocked_requests():
    patcher = mock.patch("hcloud._client.requests")
    mocked_requests = patcher.start()
    yield mocked_requests
    patcher.stop()


@pytest.fixture()
def generic_action():
    return {
        "action": {
            "id": 1,
            "command": "stop_server",
            "status": "running",
            "progress": 0,
            "started": "2016-01-30T23:50+00:00",
            "finished": None,
            "resources": [{"id": 42, "type": "server"}],
            "error": {"code": "action_failed", "message": "Action failed"},
        }
    }


@pytest.fixture()
def hetzner_client():
    client = Client(token="token")
    patcher = mock.patch.object(client, "request")
    patcher.start()
    yield client
    patcher.stop()



from __future__ import annotations

from typing import TYPE_CHECKING

from dateutil.parser import isoparse

from ..core import BaseDomain, DomainIdentityMixin

if TYPE_CHECKING:
    from ..actions import BoundAction
    from .client import BoundCertificate


class Certificate(BaseDomain, DomainIdentityMixin):
    """Certificate Domain

    :param id: int ID of Certificate
    :param name: str Name of Certificate
    :param certificate: str Certificate and chain in PEM format, in order so that each record directly certifies the one preceding
    :param not_valid_before: datetime
           Point in time when the Certificate becomes valid
    :param not_valid_after: datetime
           Point in time when the Certificate becomes invalid
    :param domain_names: List[str] List of domains and subdomains covered by this certificate
    :param fingerprint: str Fingerprint of the Certificate
    :param labels: dict
           User-defined labels (key-value pairs)
    :param created: datetime
           Point in time when the certificate was created
    :param type: str Type of Certificate
    :param status: ManagedCertificateStatus Current status of a type managed Certificate, always none for type uploaded Certificates
    """

    __slots__ = (
        "id",
        "name",
        "certificate",
        "not_valid_before",
        "not_valid_after",
        "domain_names",
        "fingerprint",
        "created",
        "labels",
        "type",
        "status",
    )
    TYPE_UPLOADED = "uploaded"
    TYPE_MANAGED = "managed"

    def __init__(
        self,
        id: int | None = None,
        name: str | None = None,
        certificate: str | None = None,
        not_valid_before: str | None = None,
        not_valid_after: str | None = None,
        domain_names: list[str] | None = None,
        fingerprint: str | None = None,
        created: str | None = None,
        labels: dict[str, str] | None = None,
        type: str | None = None,
        status: ManagedCertificateStatus | None = None,
    ):
        self.id = id
        self.name = name
        self.type = type
        self.certificate = certificate
        self.domain_names = domain_names
        self.fingerprint = fingerprint
        self.not_valid_before = isoparse(not_valid_before) if not_valid_before else None
        self.not_valid_after = isoparse(not_valid_after) if not_valid_after else None
        self.created = isoparse(created) if created else None
        self.labels = labels
        self.status = status


class ManagedCertificateStatus(BaseDomain):
    """ManagedCertificateStatus Domain

    :param issuance: str
           Status of the issuance process of the Certificate
    :param renewal: str
           Status of the renewal process of the Certificate
    :param error: ManagedCertificateError
          If issuance or renewal reports failure, this property contains information about what happened
    """

    def __init__(
        self,
        issuance: str | None = None,
        renewal: str | None = None,
        error: ManagedCertificateError | None = None,
    ):
        self.issuance = issuance
        self.renewal = renewal
        self.error = error


class ManagedCertificateError(BaseDomain):
    """ManagedCertificateError Domain

    :param code: str
        Error code identifying the error
    :param message:
        Message detailing the error
    """

    def __init__(self, code: str | None = None, message: str | None = None):
        self.code = code
        self.message = message


class CreateManagedCertificateResponse(BaseDomain):
    """Create Managed Certificate Response Domain

    :param certificate: :class:`BoundCertificate <hcloud.certificate.client.BoundCertificate>`
           The created server
    :param action: :class:`BoundAction <hcloud.actions.client.BoundAction>`
           Shows the progress of the certificate creation
    """

    __slots__ = ("certificate", "action")

    def __init__(
        self,
        certificate: BoundCertificate,
        action: BoundAction,
    ):
        self.certificate = certificate
        self.action = action



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from dateutil.parser import isoparse

from ..core import BaseDomain

if TYPE_CHECKING:
    from ..actions import BoundAction
    from ..certificates import BoundCertificate
    from ..load_balancer_types import BoundLoadBalancerType
    from ..locations import BoundLocation
    from ..metrics import Metrics
    from ..networks import BoundNetwork
    from ..servers import BoundServer
    from .client import BoundLoadBalancer


class LoadBalancer(BaseDomain):
    """LoadBalancer Domain

    :param id: int
           ID of the Load Balancer
    :param name: str
           Name of the Load Balancer (must be unique per project)
    :param created: datetime
           Point in time when the Load Balancer was created
    :param protection: dict
           Protection configuration for the Load Balancer
    :param labels: dict
            User-defined labels (key-value pairs)
    :param location: Location
            Location of the Load Balancer
    :param public_net: :class:`PublicNetwork <hcloud.load_balancers.domain.PublicNetwork>`
           Public network information.
    :param private_net: List[:class:`PrivateNet <hcloud.load_balancers.domain.PrivateNet`]
            Private networks information.
    :param algorithm: LoadBalancerAlgorithm
            The algorithm the Load Balancer is currently using
    :param services: List[LoadBalancerService]
            The services the LoadBalancer is currently serving
    :param targets: LoadBalancerTarget
            The targets the LoadBalancer is currently serving
    :param load_balancer_type: LoadBalancerType
            The type of the Load Balancer
    :param outgoing_traffic: int, None
           Outbound Traffic for the current billing period in bytes
    :param ingoing_traffic: int, None
           Inbound Traffic for the current billing period in bytes
    :param included_traffic: int
           Free Traffic for the current billing period in bytes
    """

    __slots__ = (
        "id",
        "name",
        "public_net",
        "private_net",
        "location",
        "algorithm",
        "services",
        "load_balancer_type",
        "protection",
        "labels",
        "targets",
        "created",
        "outgoing_traffic",
        "ingoing_traffic",
        "included_traffic",
    )

    # pylint: disable=too-many-locals
    def __init__(
        self,
        id: int,
        name: str | None = None,
        public_net: PublicNetwork | None = None,
        private_net: PrivateNet | None = None,
        location: BoundLocation | None = None,
        algorithm: LoadBalancerAlgorithm | None = None,
        services: list[LoadBalancerService] | None = None,
        load_balancer_type: BoundLoadBalancerType | None = None,
        protection: dict | None = None,
        labels: dict[str, str] | None = None,
        targets: list[LoadBalancerTarget] | None = None,
        created: str | None = None,
        outgoing_traffic: int | None = None,
        ingoing_traffic: int | None = None,
        included_traffic: int | None = None,
    ):
        self.id = id
        self.name = name
        self.created = isoparse(created) if created else None
        self.public_net = public_net
        self.private_net = private_net
        self.location = location
        self.algorithm = algorithm
        self.services = services
        self.load_balancer_type = load_balancer_type
        self.targets = targets
        self.protection = protection
        self.labels = labels
        self.outgoing_traffic = outgoing_traffic
        self.ingoing_traffic = ingoing_traffic
        self.included_traffic = included_traffic


class LoadBalancerService(BaseDomain):
    """LoadBalancerService Domain

    :param protocol: str
           Protocol of the service Choices: tcp, http, https
    :param listen_port: int
           Required when protocol is tcp, must be unique per Load Balancer.
    :param destination_port: int
           Required when protocol is tcp
    :param proxyprotocol: bool
            Enable proxyprotocol
    :param health_check: LoadBalancerHealthCheck
            Configuration for health checks
    :param http: LoadBalancerServiceHttp
            Configuration for http/https protocols, required when protocol is http/https
    """

    def __init__(
        self,
        protocol: str | None = None,
        listen_port: int | None = None,
        destination_port: int | None = None,
        proxyprotocol: bool | None = None,
        health_check: LoadBalancerHealthCheck | None = None,
        http: LoadBalancerServiceHttp | None = None,
    ):
        self.protocol = protocol
        self.listen_port = listen_port
        self.destination_port = destination_port
        self.proxyprotocol = proxyprotocol
        self.health_check = health_check
        self.http = http

    # pylint: disable=too-many-branches
    def to_payload(self) -> dict[str, Any]:
        """
        Generates the request payload from this domain object.
        """
        payload: dict[str, Any] = {}

        if self.protocol is not None:
            payload["protocol"] = self.protocol
        if self.listen_port is not None:
            payload["listen_port"] = self.listen_port
        if self.destination_port is not None:
            payload["destination_port"] = self.destination_port
        if self.proxyprotocol is not None:
            payload["proxyprotocol"] = self.proxyprotocol

        if self.http is not None:
            http: dict[str, Any] = {}
            if self.http.cookie_name is not None:
                http["cookie_name"] = self.http.cookie_name
            if self.http.cookie_lifetime is not None:
                http["cookie_lifetime"] = self.http.cookie_lifetime
            if self.http.redirect_http is not None:
                http["redirect_http"] = self.http.redirect_http
            if self.http.sticky_sessions is not None:
                http["sticky_sessions"] = self.http.sticky_sessions

            http["certificates"] = [
                certificate.id for certificate in self.http.certificates or []
            ]

            payload["http"] = http

        if self.health_check is not None:
            health_check: dict[str, Any] = {
                "protocol": self.health_check.protocol,
                "port": self.health_check.port,
                "interval": self.health_check.interval,
                "timeout": self.health_check.timeout,
                "retries": self.health_check.retries,
            }
            if self.health_check.protocol is not None:
                health_check["protocol"] = self.health_check.protocol
            if self.health_check.port is not None:
                health_check["port"] = self.health_check.port
            if self.health_check.interval is not None:
                health_check["interval"] = self.health_check.interval
            if self.health_check.timeout is not None:
                health_check["timeout"] = self.health_check.timeout
            if self.health_check.retries is not None:
                health_check["retries"] = self.health_check.retries

            if self.health_check.http is not None:
                health_check_http: dict[str, Any] = {}
                if self.health_check.http.domain is not None:
                    health_check_http["domain"] = self.health_check.http.domain
                if self.health_check.http.path is not None:
                    health_check_http["path"] = self.health_check.http.path
                if self.health_check.http.response is not None:
                    health_check_http["response"] = self.health_check.http.response
                if self.health_check.http.status_codes is not None:
                    health_check_http[
                        "status_codes"
                    ] = self.health_check.http.status_codes
                if self.health_check.http.tls is not None:
                    health_check_http["tls"] = self.health_check.http.tls

                health_check["http"] = health_check_http

            payload["health_check"] = health_check
        return payload


class LoadBalancerServiceHttp(BaseDomain):
    """LoadBalancerServiceHttp Domain

    :param cookie_name: str
        Name of the cookie used for Session Stickness
    :param cookie_lifetime: str
        Lifetime of the cookie used for Session Stickness
    :param certificates: list
            IDs of the Certificates to use for TLS/SSL termination by the Load Balancer; empty for TLS/SSL passthrough or if protocol is "http"
    :param redirect_http: bool
           Redirect traffic from http port 80 to port 443
    :param sticky_sessions: bool
           Use sticky sessions. Only available if protocol is "http" or "https".
    """

    def __init__(
        self,
        cookie_name: str | None = None,
        cookie_lifetime: str | None = None,
        certificates: list[BoundCertificate] | None = None,
        redirect_http: bool | None = None,
        sticky_sessions: bool | None = None,
    ):
        self.cookie_name = cookie_name
        self.cookie_lifetime = cookie_lifetime
        self.certificates = certificates
        self.redirect_http = redirect_http
        self.sticky_sessions = sticky_sessions


class LoadBalancerHealthCheck(BaseDomain):
    """LoadBalancerHealthCheck Domain

    :param protocol: str
        Protocol of the service Choices: tcp, http, https
    :param port: int
        Port the healthcheck will be performed on
    :param interval: int
           Interval we trigger health check in
    :param timeout: int
            Timeout in sec after a try is assumed as timeout
    :param retries: int
            Retries we perform until we assume a target as unhealthy
    :param http: LoadBalancerHealtCheckHttp
            HTTP Config
    """

    def __init__(
        self,
        protocol: str | None = None,
        port: int | None = None,
        interval: int | None = None,
        timeout: int | None = None,
        retries: int | None = None,
        http: LoadBalancerHealtCheckHttp | None = None,
    ):
        self.protocol = protocol
        self.port = port
        self.interval = interval
        self.timeout = timeout
        self.retries = retries
        self.http = http


class LoadBalancerHealtCheckHttp(BaseDomain):
    """LoadBalancerHealtCheckHttp Domain

    :param domain: str
            Domain name to send in HTTP request. Can be null: In that case we will not send a domain name
    :param path: str
            HTTP Path send in Request
    :param response: str
            Optional HTTP response to receive in order to pass the health check
    :param status_codes: list
            List of HTTP status codes to receive in order to pass the health check
    :param tls: bool
            Type of health check
    """

    def __init__(
        self,
        domain: str | None = None,
        path: str | None = None,
        response: str | None = None,
        status_codes: list | None = None,
        tls: bool | None = None,
    ):
        self.domain = domain
        self.path = path
        self.response = response
        self.status_codes = status_codes
        self.tls = tls


class LoadBalancerTarget(BaseDomain):
    """LoadBalancerTarget Domain

    :param type: str
            Type of the resource, can be server or label_selector
    :param server: Server
            Target server
    :param label_selector: LoadBalancerTargetLabelSelector
            Target label selector
    :param ip: LoadBalancerTargetIP
            Target IP
    :param use_private_ip: bool
            use the private IP instead of primary public IP
    :param health_status: list
            List of health statuses of the services on this target. Only present for target types "server" and "ip".
    """

    def __init__(
        self,
        type: str | None = None,
        server: BoundServer | None = None,
        label_selector: LoadBalancerTargetLabelSelector | None = None,
        ip: LoadBalancerTargetIP | None = None,
        use_private_ip: bool | None = None,
        health_status: list[LoadBalancerTargetHealthStatus] | None = None,
    ):
        self.type = type
        self.server = server
        self.label_selector = label_selector
        self.ip = ip
        self.use_private_ip = use_private_ip
        self.health_status = health_status

    def to_payload(self) -> dict[str, Any]:
        """
        Generates the request payload from this domain object.
        """
        payload: dict[str, Any] = {
            "type": self.type,
        }
        if self.use_private_ip is not None:
            payload["use_private_ip"] = self.use_private_ip

        if self.type == "server":
            if self.server is None:
                raise ValueError(f"server is not defined in target {self!r}")
            payload["server"] = {"id": self.server.id}

        elif self.type == "label_selector":
            if self.label_selector is None:
                raise ValueError(f"label_selector is not defined in target {self!r}")
            payload["label_selector"] = {"selector": self.label_selector.selector}

        elif self.type == "ip":
            if self.ip is None:
                raise ValueError(f"ip is not defined in target {self!r}")
            payload["ip"] = {"ip": self.ip.ip}

        return payload


class LoadBalancerTargetHealthStatus(BaseDomain):
    """LoadBalancerTargetHealthStatus Domain

    :param listen_port: Load Balancer Target listen port
    :param status: Load Balancer Target status. Choices: healthy, unhealthy, unknown
    """

    def __init__(
        self,
        listen_port: int | None = None,
        status: str | None = None,
    ):
        self.listen_port = listen_port
        self.status = status


class LoadBalancerTargetLabelSelector(BaseDomain):
    """LoadBalancerTargetLabelSelector Domain

    :param selector: str Target label selector
    """

    def __init__(self, selector: str | None = None):
        self.selector = selector


class LoadBalancerTargetIP(BaseDomain):
    """LoadBalancerTargetIP Domain

    :param ip: str Target IP
    """

    def __init__(self, ip: str | None = None):
        self.ip = ip


class LoadBalancerAlgorithm(BaseDomain):
    """LoadBalancerAlgorithm Domain

    :param type: str
            Algorithm of the Load Balancer. Choices: round_robin, least_connections
    """

    def __init__(self, type: str | None = None):
        self.type = type


class PublicNetwork(BaseDomain):
    """Public Network Domain

    :param ipv4: :class:`IPv4Address <hcloud.load_balancers.domain.IPv4Address>`
    :param ipv6: :class:`IPv6Network <hcloud.load_balancers.domain.IPv6Network>`
    :param enabled:  boolean
    """

    __slots__ = ("ipv4", "ipv6", "enabled")

    def __init__(
        self,
        ipv4: IPv4Address,
        ipv6: IPv6Network,
        enabled: bool,
    ):
        self.ipv4 = ipv4
        self.ipv6 = ipv6
        self.enabled = enabled


class IPv4Address(BaseDomain):
    """IPv4 Address Domain

    :param ip: str
           The IPv4 Address
    """

    __slots__ = ("ip", "dns_ptr")

    def __init__(
        self,
        ip: str,
        dns_ptr: str,
    ):
        self.ip = ip
        self.dns_ptr = dns_ptr


class IPv6Network(BaseDomain):
    """IPv6 Network Domain

    :param ip: str
           The IPv6 Network as CIDR Notation
    """

    __slots__ = ("ip", "dns_ptr")

    def __init__(
        self,
        ip: str,
        dns_ptr: str,
    ):
        self.ip = ip
        self.dns_ptr = dns_ptr


class PrivateNet(BaseDomain):
    """PrivateNet Domain

    :param network: :class:`BoundNetwork <hcloud.networks.client.BoundNetwork>`
           The Network the LoadBalancer is attached to
    :param ip: str
           The main IP Address of the LoadBalancer in the Network
    """

    __slots__ = ("network", "ip")

    def __init__(
        self,
        network: BoundNetwork,
        ip: str,
    ):
        self.network = network
        self.ip = ip


class CreateLoadBalancerResponse(BaseDomain):
    """Create Load Balancer Response Domain

    :param load_balancer: :class:`BoundLoadBalancer <hcloud.load_balancers.client.BoundLoadBalancer>`
           The created Load Balancer
    :param action: :class:`BoundAction <hcloud.actions.client.BoundAction>`
           Shows the progress of the Load Balancer creation
    """

    __slots__ = ("load_balancer", "action")

    def __init__(
        self,
        load_balancer: BoundLoadBalancer,
        action: BoundAction,
    ):
        self.load_balancer = load_balancer
        self.action = action


MetricsType = Literal[
    "open_connections",
    "connections_per_second",
    "requests_per_second",
    "bandwidth",
]


class GetMetricsResponse(BaseDomain):
    """Get a Load Balancer Metrics Response Domain

    :param metrics: The Load Balancer metrics
    """

    __slots__ = ("metrics",)

    def __init__(
        self,
        metrics: Metrics,
    ):
        self.metrics = metrics



from __future__ import annotations

import pytest


@pytest.fixture()
def volume_response():
    return {
        "volume": {
            "id": 1,
            "created": "2016-01-30T23:50:11+00:00",
            "name": "database-storage",
            "server": 12,
            "location": {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
            },
            "size": 42,
            "linux_device": "/dev/disk/by-id/scsi-0HC_Volume_4711",
            "protection": {"delete": False},
            "format": "xfs",
            "labels": {},
            "status": "available",
        }
    }


@pytest.fixture()
def two_volumes_response():
    return {
        "volumes": [
            {
                "id": 1,
                "created": "2016-01-30T23:50:11+00:00",
                "name": "database-storage",
                "server": 12,
                "location": {
                    "id": 1,
                    "name": "fsn1",
                    "description": "Falkenstein DC Park 1",
                    "country": "DE",
                    "city": "Falkenstein",
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                },
                "size": 42,
                "linux_device": "/dev/disk/by-id/scsi-0HC_Volume_4711",
                "protection": {"delete": False},
                "format": "xfs",
                "labels": {},
                "status": "available",
            },
            {
                "id": 2,
                "created": "2016-01-30T23:50:11+00:00",
                "name": "vault-storage",
                "server": 10,
                "location": {
                    "id": 1,
                    "name": "fsn1",
                    "description": "Falkenstein DC Park 2",
                    "country": "DE",
                    "city": "Falkenstein",
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                },
                "size": 42,
                "linux_device": "/dev/disk/by-id/scsi-0HC_Volume_4711",
                "protection": {"delete": False},
                "format": "xfs",
                "labels": {},
                "status": "available",
            },
        ]
    }


@pytest.fixture()
def one_volumes_response():
    return {
        "volumes": [
            {
                "id": 1,
                "created": "2016-01-30T23:50:11+00:00",
                "name": "database-storage",
                "server": 12,
                "location": {
                    "id": 1,
                    "name": "fsn1",
                    "description": "Falkenstein DC Park 1",
                    "country": "DE",
                    "city": "Falkenstein",
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                },
                "size": 42,
                "linux_device": "/dev/disk/by-id/scsi-0HC_Volume_4711",
                "protection": {"delete": False},
                "format": "xfs",
                "labels": {},
                "status": "available",
            }
        ]
    }


@pytest.fixture()
def volume_create_response():
    return {
        "volume": {
            "id": 4711,
            "created": "2016-01-30T23:50:11+00:00",
            "name": "database-storage",
            "server": 12,
            "location": {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
            },
            "size": 42,
            "linux_device": "/dev/disk/by-id/scsi-0HC_Volume_4711",
            "protection": {"delete": False},
            "format": "xfs",
            "labels": {},
            "status": "available",
        },
        "action": {
            "id": 13,
            "command": "create_volume",
            "status": "running",
            "progress": 0,
            "started": "2016-01-30T23:50+00:00",
            "finished": None,
            "resources": [{"id": 42, "type": "server"}],
            "error": {"code": "action_failed", "message": "Action failed"},
        },
        "next_actions": [
            {
                "id": 13,
                "command": "start_server",
                "status": "running",
                "progress": 0,
                "started": "2016-01-30T23:50+00:00",
                "finished": None,
                "resources": [{"id": 42, "type": "server"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            }
        ],
    }


@pytest.fixture()
def response_update_volume():
    return {
        "volume": {
            "id": 4711,
            "created": "2016-01-30T23:50:11+00:00",
            "name": "new-name",
            "server": 12,
            "location": {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
            },
            "format": "xfs",
            "size": 42,
            "linux_device": "/dev/disk/by-id/scsi-0HC_Volume_4711",
            "protection": {"delete": False},
            "labels": {},
            "status": "available",
        }
    }


@pytest.fixture()
def response_get_actions():
    return {
        "actions": [
            {
                "id": 13,
                "command": "attach_volume",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [{"id": 42, "type": "server"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            }
        ]
    }



from __future__ import annotations

from ._client import Client as Client  # noqa pylint: disable=C0414
from ._exceptions import (  # noqa pylint: disable=C0414
    APIException as APIException,
    HCloudException as HCloudException,
)



from __future__ import annotations

import pytest


@pytest.fixture()
def load_balancer_type_response():
    return {
        "load_balancer_type": {
            "id": 1,
            "name": "LB11",
            "description": "LB11",
            "max_connections": 1,
            "max_services": 1,
            "max_targets": 1,
            "max_assigned_certificates": 1,
            "deprecated": None,
            "prices": [
                {
                    "location": "fsn1",
                    "price_hourly": {
                        "net": "1.0000000000",
                        "gross": "1.1900000000000000",
                    },
                    "price_monthly": {
                        "net": "1.0000000000",
                        "gross": "1.1900000000000000",
                    },
                }
            ],
        }
    }


@pytest.fixture()
def two_load_balancer_types_response():
    return {
        "load_balancer_types": [
            {
                "id": 1,
                "name": "LB11",
                "description": "LB11D",
                "max_connections": 1,
                "max_services": 1,
                "max_targets": 1,
                "max_assigned_certificates": 1,
                "deprecated": None,
                "prices": [
                    {
                        "location": "fsn1",
                        "price_hourly": {
                            "net": "1.0000000000",
                            "gross": "1.1900000000000000",
                        },
                        "price_monthly": {
                            "net": "1.0000000000",
                            "gross": "1.1900000000000000",
                        },
                    }
                ],
            },
            {
                "id": 2,
                "name": "LB21",
                "description": "LB21D",
                "max_connections": 2,
                "max_services": 2,
                "max_targets": 2,
                "max_assigned_certificates": 2,
                "deprecated": None,
                "prices": [
                    {
                        "location": "fsn1",
                        "price_hourly": {
                            "net": "1.0000000000",
                            "gross": "1.1900000000000000",
                        },
                        "price_monthly": {
                            "net": "1.0000000000",
                            "gross": "1.1900000000000000",
                        },
                    }
                ],
            },
        ]
    }


@pytest.fixture()
def one_load_balancer_types_response():
    return {
        "load_balancer_types": [
            {
                "id": 2,
                "name": "LB21",
                "description": "LB21D",
                "max_connections": 2,
                "max_services": 2,
                "max_targets": 2,
                "max_assigned_certificates": 2,
                "deprecated": None,
                "prices": [
                    {
                        "location": "fsn1",
                        "price_hourly": {
                            "net": "1.0000000000",
                            "gross": "1.1900000000000000",
                        },
                        "price_monthly": {
                            "net": "1.0000000000",
                            "gross": "1.1900000000000000",
                        },
                    }
                ],
            }
        ]
    }



from __future__ import annotations

from datetime import datetime, timezone

import pytest

from hcloud.isos import Iso


class TestIso:
    @pytest.fixture()
    def deprecated_iso(self):
        return Iso(
            **{
                "id": 10433,
                "name": "vyos-1.4-rolling-202111150317-amd64.iso",
                "description": "VyOS 1.4 (amd64)",
                "type": "public",
                "deprecation": {
                    "announced": "2023-10-05T08:27:01Z",
                    "unavailable_after": "2023-11-05T08:27:01Z",
                },
                "architecture": "x86",
                "deprecated": "2023-11-05T08:27:01Z",
            }
        )

    def test_deprecation(self, deprecated_iso: Iso):
        with pytest.deprecated_call():
            assert deprecated_iso.deprecated == datetime(
                2023, 11, 5, 8, 27, 1, tzinfo=timezone.utc
            )
        assert deprecated_iso.deprecation is not None
        assert deprecated_iso.deprecation.announced == datetime(
            2023, 10, 5, 8, 27, 1, tzinfo=timezone.utc
        )
        assert deprecated_iso.deprecation.unavailable_after == datetime(
            2023, 11, 5, 8, 27, 1, tzinfo=timezone.utc
        )



from __future__ import annotations

from unittest import mock

import pytest

from hcloud.ssh_keys import BoundSSHKey, SSHKey, SSHKeysClient


class TestBoundSSHKey:
    @pytest.fixture()
    def bound_ssh_key(self, hetzner_client):
        return BoundSSHKey(client=hetzner_client.ssh_keys, data=dict(id=14))

    def test_bound_ssh_key_init(self, ssh_key_response):
        bound_ssh_key = BoundSSHKey(
            client=mock.MagicMock(), data=ssh_key_response["ssh_key"]
        )

        assert bound_ssh_key.id == 2323
        assert bound_ssh_key.name == "My ssh key"
        assert (
            bound_ssh_key.fingerprint
            == "b7:2f:30:a0:2f:6c:58:6c:21:04:58:61:ba:06:3b:2f"
        )
        assert bound_ssh_key.public_key == "ssh-rsa AAAjjk76kgf...Xt"

    def test_update(self, hetzner_client, bound_ssh_key, response_update_ssh_key):
        hetzner_client.request.return_value = response_update_ssh_key
        ssh_key = bound_ssh_key.update(name="New name")
        hetzner_client.request.assert_called_with(
            url="/ssh_keys/14", method="PUT", json={"name": "New name"}
        )

        assert ssh_key.id == 2323
        assert ssh_key.name == "New name"

    def test_delete(self, hetzner_client, bound_ssh_key, generic_action):
        hetzner_client.request.return_value = generic_action
        delete_success = bound_ssh_key.delete()
        hetzner_client.request.assert_called_with(url="/ssh_keys/14", method="DELETE")

        assert delete_success is True


class TestSSHKeysClient:
    @pytest.fixture()
    def ssh_keys_client(self):
        return SSHKeysClient(client=mock.MagicMock())

    def test_get_by_id(self, ssh_keys_client, ssh_key_response):
        ssh_keys_client._client.request.return_value = ssh_key_response
        ssh_key = ssh_keys_client.get_by_id(1)
        ssh_keys_client._client.request.assert_called_with(
            url="/ssh_keys/1", method="GET"
        )
        assert ssh_key._client is ssh_keys_client
        assert ssh_key.id == 2323
        assert ssh_key.name == "My ssh key"

    @pytest.mark.parametrize(
        "params",
        [
            {
                "name": "My ssh key",
                "fingerprint": "b7:2f:30:a0:2f:6c:58:6c:21:04:58:61:ba:06:3b:2f",
                "label_selector": "k==v",
                "page": 1,
                "per_page": 10,
            },
            {"name": ""},
            {},
        ],
    )
    def test_get_list(self, ssh_keys_client, two_ssh_keys_response, params):
        ssh_keys_client._client.request.return_value = two_ssh_keys_response
        result = ssh_keys_client.get_list(**params)
        ssh_keys_client._client.request.assert_called_with(
            url="/ssh_keys", method="GET", params=params
        )

        ssh_keys = result.ssh_keys
        assert len(ssh_keys) == 2

        ssh_keys1 = ssh_keys[0]
        ssh_keys2 = ssh_keys[1]

        assert ssh_keys1._client is ssh_keys_client
        assert ssh_keys1.id == 2323
        assert ssh_keys1.name == "SSH-Key"

        assert ssh_keys2._client is ssh_keys_client
        assert ssh_keys2.id == 2324
        assert ssh_keys2.name == "SSH-Key"

    @pytest.mark.parametrize(
        "params", [{"name": "My ssh key", "label_selector": "label1"}, {}]
    )
    def test_get_all(self, ssh_keys_client, two_ssh_keys_response, params):
        ssh_keys_client._client.request.return_value = two_ssh_keys_response
        ssh_keys = ssh_keys_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})
        ssh_keys_client._client.request.assert_called_with(
            url="/ssh_keys", method="GET", params=params
        )

        assert len(ssh_keys) == 2

        ssh_keys1 = ssh_keys[0]
        ssh_keys2 = ssh_keys[1]

        assert ssh_keys1._client is ssh_keys_client
        assert ssh_keys1.id == 2323
        assert ssh_keys1.name == "SSH-Key"

        assert ssh_keys2._client is ssh_keys_client
        assert ssh_keys2.id == 2324
        assert ssh_keys2.name == "SSH-Key"

    def test_get_by_name(self, ssh_keys_client, one_ssh_keys_response):
        ssh_keys_client._client.request.return_value = one_ssh_keys_response
        ssh_keys = ssh_keys_client.get_by_name("SSH-Key")

        params = {"name": "SSH-Key"}
        ssh_keys_client._client.request.assert_called_with(
            url="/ssh_keys", method="GET", params=params
        )

        assert ssh_keys._client is ssh_keys_client
        assert ssh_keys.id == 2323
        assert ssh_keys.name == "SSH-Key"

    def test_get_by_fingerprint(self, ssh_keys_client, one_ssh_keys_response):
        ssh_keys_client._client.request.return_value = one_ssh_keys_response
        ssh_keys = ssh_keys_client.get_by_fingerprint(
            "b7:2f:30:a0:2f:6c:58:6c:21:04:58:61:ba:06:3b:2f"
        )

        params = {"fingerprint": "b7:2f:30:a0:2f:6c:58:6c:21:04:58:61:ba:06:3b:2f"}
        ssh_keys_client._client.request.assert_called_with(
            url="/ssh_keys", method="GET", params=params
        )

        assert ssh_keys._client is ssh_keys_client
        assert ssh_keys.id == 2323
        assert ssh_keys.name == "SSH-Key"

    def test_create(self, ssh_keys_client, ssh_key_response):
        ssh_keys_client._client.request.return_value = ssh_key_response
        ssh_key = ssh_keys_client.create(
            name="My ssh key", public_key="ssh-rsa AAAjjk76kgf...Xt"
        )
        ssh_keys_client._client.request.assert_called_with(
            url="/ssh_keys",
            method="POST",
            json={"name": "My ssh key", "public_key": "ssh-rsa AAAjjk76kgf...Xt"},
        )

        assert ssh_key.id == 2323
        assert ssh_key.name == "My ssh key"

    @pytest.mark.parametrize(
        "ssh_key", [SSHKey(id=1), BoundSSHKey(mock.MagicMock(), dict(id=1))]
    )
    def test_update(self, ssh_keys_client, ssh_key, response_update_ssh_key):
        ssh_keys_client._client.request.return_value = response_update_ssh_key
        ssh_key = ssh_keys_client.update(ssh_key, name="New name")
        ssh_keys_client._client.request.assert_called_with(
            url="/ssh_keys/1", method="PUT", json={"name": "New name"}
        )

        assert ssh_key.id == 2323
        assert ssh_key.name == "New name"

    @pytest.mark.parametrize(
        "ssh_key", [SSHKey(id=1), BoundSSHKey(mock.MagicMock(), dict(id=1))]
    )
    def test_delete(self, ssh_keys_client, ssh_key, generic_action):
        ssh_keys_client._client.request.return_value = generic_action
        delete_success = ssh_keys_client.delete(ssh_key)
        ssh_keys_client._client.request.assert_called_with(
            url="/ssh_keys/1", method="DELETE"
        )

        assert delete_success is True



from __future__ import annotations

from unittest import mock

import pytest

from hcloud.actions import BoundAction
from hcloud.firewalls import (
    BoundFirewall,
    Firewall,
    FirewallResource,
    FirewallResourceLabelSelector,
    FirewallRule,
    FirewallsClient,
)
from hcloud.servers import Server


class TestBoundFirewall:
    @pytest.fixture()
    def bound_firewall(self, hetzner_client):
        return BoundFirewall(client=hetzner_client.firewalls, data=dict(id=1))

    def test_bound_firewall_init(self, firewall_response):
        bound_firewall = BoundFirewall(
            client=mock.MagicMock(), data=firewall_response["firewall"]
        )

        assert bound_firewall.id == 38
        assert bound_firewall.name == "Corporate Intranet Protection"
        assert bound_firewall.labels == {}
        assert isinstance(bound_firewall.rules, list)
        assert len(bound_firewall.rules) == 2

        assert isinstance(bound_firewall.applied_to, list)
        assert len(bound_firewall.applied_to) == 2
        assert bound_firewall.applied_to[0].server.id == 42
        assert bound_firewall.applied_to[0].type == "server"
        assert bound_firewall.applied_to[1].label_selector.selector == "key==value"
        assert bound_firewall.applied_to[1].type == "label_selector"

        firewall_in_rule = bound_firewall.rules[0]
        assert isinstance(firewall_in_rule, FirewallRule)
        assert firewall_in_rule.direction == FirewallRule.DIRECTION_IN
        assert firewall_in_rule.protocol == FirewallRule.PROTOCOL_TCP
        assert firewall_in_rule.port == "80"
        assert isinstance(firewall_in_rule.source_ips, list)
        assert len(firewall_in_rule.source_ips) == 3
        assert firewall_in_rule.source_ips == [
            "28.239.13.1/32",
            "28.239.14.0/24",
            "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
        ]
        assert isinstance(firewall_in_rule.destination_ips, list)
        assert len(firewall_in_rule.destination_ips) == 0
        assert firewall_in_rule.description == "allow http in"

        firewall_out_rule = bound_firewall.rules[1]
        assert isinstance(firewall_out_rule, FirewallRule)
        assert firewall_out_rule.direction == FirewallRule.DIRECTION_OUT
        assert firewall_out_rule.protocol == FirewallRule.PROTOCOL_TCP
        assert firewall_out_rule.port == "80"
        assert isinstance(firewall_out_rule.source_ips, list)
        assert len(firewall_out_rule.source_ips) == 0
        assert isinstance(firewall_out_rule.destination_ips, list)
        assert len(firewall_out_rule.destination_ips) == 3
        assert firewall_out_rule.destination_ips == [
            "28.239.13.1/32",
            "28.239.14.0/24",
            "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
        ]
        assert firewall_out_rule.description == "allow http out"

    @pytest.mark.parametrize(
        "params", [{}, {"sort": ["created"], "page": 1, "per_page": 2}]
    )
    def test_get_actions_list(
        self, hetzner_client, bound_firewall, response_get_actions, params
    ):
        hetzner_client.request.return_value = response_get_actions
        result = bound_firewall.get_actions_list(**params)
        hetzner_client.request.assert_called_with(
            url="/firewalls/1/actions", method="GET", params=params
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == hetzner_client.actions
        assert actions[0].id == 13
        assert actions[0].command == "set_firewall_rules"

    @pytest.mark.parametrize("params", [{}, {"sort": ["created"]}])
    def test_get_actions(
        self, hetzner_client, bound_firewall, response_get_actions, params
    ):
        hetzner_client.request.return_value = response_get_actions
        actions = bound_firewall.get_actions(**params)

        params.update({"page": 1, "per_page": 50})

        hetzner_client.request.assert_called_with(
            url="/firewalls/1/actions", method="GET", params=params
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == hetzner_client.actions
        assert actions[0].id == 13
        assert actions[0].command == "set_firewall_rules"

    def test_update(self, hetzner_client, bound_firewall, response_update_firewall):
        hetzner_client.request.return_value = response_update_firewall
        firewall = bound_firewall.update(
            name="New Corporate Intranet Protection", labels={}
        )
        hetzner_client.request.assert_called_with(
            url="/firewalls/1",
            method="PUT",
            json={"name": "New Corporate Intranet Protection", "labels": {}},
        )

        assert firewall.id == 38
        assert firewall.name == "New Corporate Intranet Protection"

    def test_delete(self, hetzner_client, bound_firewall):
        delete_success = bound_firewall.delete()
        hetzner_client.request.assert_called_with(url="/firewalls/1", method="DELETE")

        assert delete_success is True

    def test_set_rules(self, hetzner_client, bound_firewall, response_set_rules):
        hetzner_client.request.return_value = response_set_rules
        actions = bound_firewall.set_rules(
            [
                FirewallRule(
                    direction=FirewallRule.DIRECTION_IN,
                    protocol=FirewallRule.PROTOCOL_ICMP,
                    source_ips=["0.0.0.0/0", "::/0"],
                    description="New firewall description",
                )
            ]
        )
        hetzner_client.request.assert_called_with(
            url="/firewalls/1/actions/set_rules",
            method="POST",
            json={
                "rules": [
                    {
                        "direction": "in",
                        "protocol": "icmp",
                        "source_ips": ["0.0.0.0/0", "::/0"],
                        "description": "New firewall description",
                    }
                ]
            },
        )

        assert actions[0].id == 13
        assert actions[0].progress == 100

    def test_apply_to_resources(
        self, hetzner_client, bound_firewall, response_set_rules
    ):
        hetzner_client.request.return_value = response_set_rules
        actions = bound_firewall.apply_to_resources(
            [FirewallResource(type=FirewallResource.TYPE_SERVER, server=Server(id=5))]
        )
        hetzner_client.request.assert_called_with(
            url="/firewalls/1/actions/apply_to_resources",
            method="POST",
            json={"apply_to": [{"type": "server", "server": {"id": 5}}]},
        )

        assert actions[0].id == 13
        assert actions[0].progress == 100

    def test_remove_from_resources(
        self, hetzner_client, bound_firewall, response_set_rules
    ):
        hetzner_client.request.return_value = response_set_rules
        actions = bound_firewall.remove_from_resources(
            [FirewallResource(type=FirewallResource.TYPE_SERVER, server=Server(id=5))]
        )
        hetzner_client.request.assert_called_with(
            url="/firewalls/1/actions/remove_from_resources",
            method="POST",
            json={"remove_from": [{"type": "server", "server": {"id": 5}}]},
        )

        assert actions[0].id == 13
        assert actions[0].progress == 100


class TestFirewallsClient:
    @pytest.fixture()
    def firewalls_client(self):
        return FirewallsClient(client=mock.MagicMock())

    def test_get_by_id(self, firewalls_client, firewall_response):
        firewalls_client._client.request.return_value = firewall_response
        firewall = firewalls_client.get_by_id(1)
        firewalls_client._client.request.assert_called_with(
            url="/firewalls/1", method="GET"
        )
        assert firewall._client is firewalls_client
        assert firewall.id == 38
        assert firewall.name == "Corporate Intranet Protection"

    @pytest.mark.parametrize(
        "params",
        [
            {
                "name": "Corporate Intranet Protection",
                "sort": "id",
                "label_selector": "k==v",
                "page": 1,
                "per_page": 10,
            },
            {"name": ""},
            {},
        ],
    )
    def test_get_list(self, firewalls_client, two_firewalls_response, params):
        firewalls_client._client.request.return_value = two_firewalls_response
        result = firewalls_client.get_list(**params)
        firewalls_client._client.request.assert_called_with(
            url="/firewalls", method="GET", params=params
        )

        firewalls = result.firewalls
        assert result.meta is None

        assert len(firewalls) == 2

        firewalls1 = firewalls[0]
        firewalls2 = firewalls[1]

        assert firewalls1._client is firewalls_client
        assert firewalls1.id == 38
        assert firewalls1.name == "Corporate Intranet Protection"

        assert firewalls2._client is firewalls_client
        assert firewalls2.id == 39
        assert firewalls2.name == "Corporate Extranet Protection"

    @pytest.mark.parametrize(
        "params",
        [
            {
                "name": "Corporate Intranet Protection",
                "sort": "id",
                "label_selector": "k==v",
            },
            {},
        ],
    )
    def test_get_all(self, firewalls_client, two_firewalls_response, params):
        firewalls_client._client.request.return_value = two_firewalls_response
        firewalls = firewalls_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        firewalls_client._client.request.assert_called_with(
            url="/firewalls", method="GET", params=params
        )

        assert len(firewalls) == 2

        firewalls1 = firewalls[0]
        firewalls2 = firewalls[1]

        assert firewalls1._client is firewalls_client
        assert firewalls1.id == 38
        assert firewalls1.name == "Corporate Intranet Protection"

        assert firewalls2._client is firewalls_client
        assert firewalls2.id == 39
        assert firewalls2.name == "Corporate Extranet Protection"

    def test_get_by_name(self, firewalls_client, one_firewalls_response):
        firewalls_client._client.request.return_value = one_firewalls_response
        firewall = firewalls_client.get_by_name("Corporate Intranet Protection")

        params = {"name": "Corporate Intranet Protection"}

        firewalls_client._client.request.assert_called_with(
            url="/firewalls", method="GET", params=params
        )

        assert firewall._client is firewalls_client
        assert firewall.id == 38
        assert firewall.name == "Corporate Intranet Protection"

    @pytest.mark.parametrize(
        "firewall", [Firewall(id=1), BoundFirewall(mock.MagicMock(), dict(id=1))]
    )
    def test_get_actions_list(self, firewalls_client, firewall, response_get_actions):
        firewalls_client._client.request.return_value = response_get_actions
        result = firewalls_client.get_actions_list(firewall)
        firewalls_client._client.request.assert_called_with(
            url="/firewalls/1/actions", method="GET", params={}
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)

        assert actions[0]._client == firewalls_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "set_firewall_rules"

    def test_create(self, firewalls_client, response_create_firewall):
        firewalls_client._client.request.return_value = response_create_firewall
        response = firewalls_client.create(
            "Corporate Intranet Protection",
            rules=[
                FirewallRule(
                    direction=FirewallRule.DIRECTION_IN,
                    protocol=FirewallRule.PROTOCOL_ICMP,
                    source_ips=["0.0.0.0/0"],
                )
            ],
            resources=[
                FirewallResource(
                    type=FirewallResource.TYPE_SERVER, server=Server(id=4711)
                ),
                FirewallResource(
                    type=FirewallResource.TYPE_LABEL_SELECTOR,
                    label_selector=FirewallResourceLabelSelector(selector="key==value"),
                ),
            ],
        )
        firewalls_client._client.request.assert_called_with(
            url="/firewalls",
            method="POST",
            json={
                "name": "Corporate Intranet Protection",
                "rules": [
                    {"direction": "in", "protocol": "icmp", "source_ips": ["0.0.0.0/0"]}
                ],
                "apply_to": [
                    {"type": "server", "server": {"id": 4711}},
                    {
                        "type": "label_selector",
                        "label_selector": {"selector": "key==value"},
                    },
                ],
            },
        )

        bound_firewall = response.firewall
        actions = response.actions

        assert bound_firewall._client is firewalls_client
        assert bound_firewall.id == 38
        assert bound_firewall.name == "Corporate Intranet Protection"
        assert len(bound_firewall.applied_to) == 2

        assert len(actions) == 2

    @pytest.mark.parametrize(
        "firewall", [Firewall(id=38), BoundFirewall(mock.MagicMock(), dict(id=38))]
    )
    def test_update(self, firewalls_client, firewall, response_update_firewall):
        firewalls_client._client.request.return_value = response_update_firewall
        firewall = firewalls_client.update(
            firewall, name="New Corporate Intranet Protection", labels={}
        )
        firewalls_client._client.request.assert_called_with(
            url="/firewalls/38",
            method="PUT",
            json={"name": "New Corporate Intranet Protection", "labels": {}},
        )

        assert firewall.id == 38
        assert firewall.name == "New Corporate Intranet Protection"

    @pytest.mark.parametrize(
        "firewall", [Firewall(id=1), BoundFirewall(mock.MagicMock(), dict(id=1))]
    )
    def test_set_rules(self, firewalls_client, firewall, response_set_rules):
        firewalls_client._client.request.return_value = response_set_rules
        actions = firewalls_client.set_rules(
            firewall,
            [
                FirewallRule(
                    direction=FirewallRule.DIRECTION_IN,
                    protocol=FirewallRule.PROTOCOL_ICMP,
                    source_ips=["0.0.0.0/0", "::/0"],
                )
            ],
        )
        firewalls_client._client.request.assert_called_with(
            url="/firewalls/1/actions/set_rules",
            method="POST",
            json={
                "rules": [
                    {
                        "direction": "in",
                        "protocol": "icmp",
                        "source_ips": ["0.0.0.0/0", "::/0"],
                    }
                ]
            },
        )

        assert actions[0].id == 13
        assert actions[0].progress == 100

    @pytest.mark.parametrize(
        "firewall", [Firewall(id=1), BoundFirewall(mock.MagicMock(), dict(id=1))]
    )
    def test_delete(self, firewalls_client, firewall):
        delete_success = firewalls_client.delete(firewall)
        firewalls_client._client.request.assert_called_with(
            url="/firewalls/1", method="DELETE"
        )

        assert delete_success is True

    @pytest.mark.parametrize(
        "firewall", [Firewall(id=1), BoundFirewall(mock.MagicMock(), dict(id=1))]
    )
    def test_apply_to_resources(self, firewalls_client, firewall, response_set_rules):
        firewalls_client._client.request.return_value = response_set_rules

        actions = firewalls_client.apply_to_resources(
            firewall,
            [FirewallResource(type=FirewallResource.TYPE_SERVER, server=Server(id=5))],
        )
        firewalls_client._client.request.assert_called_with(
            url="/firewalls/1/actions/apply_to_resources",
            method="POST",
            json={"apply_to": [{"type": "server", "server": {"id": 5}}]},
        )

        assert actions[0].id == 13
        assert actions[0].progress == 100

    @pytest.mark.parametrize(
        "firewall", [Firewall(id=1), BoundFirewall(mock.MagicMock(), dict(id=1))]
    )
    def test_remove_from_resources(
        self, firewalls_client, firewall, response_set_rules
    ):
        firewalls_client._client.request.return_value = response_set_rules

        actions = firewalls_client.remove_from_resources(
            firewall,
            [FirewallResource(type=FirewallResource.TYPE_SERVER, server=Server(id=5))],
        )
        firewalls_client._client.request.assert_called_with(
            url="/firewalls/1/actions/remove_from_resources",
            method="POST",
            json={"remove_from": [{"type": "server", "server": {"id": 5}}]},
        )

        assert actions[0].id == 13
        assert actions[0].progress == 100

    def test_actions_get_by_id(self, firewalls_client, response_get_actions):
        firewalls_client._client.request.return_value = {
            "action": response_get_actions["actions"][0]
        }
        action = firewalls_client.actions.get_by_id(13)

        firewalls_client._client.request.assert_called_with(
            url="/firewalls/actions/13", method="GET"
        )

        assert isinstance(action, BoundAction)
        assert action._client == firewalls_client._client.actions
        assert action.id == 13
        assert action.command == "set_firewall_rules"

    def test_actions_get_list(self, firewalls_client, response_get_actions):
        firewalls_client._client.request.return_value = response_get_actions
        result = firewalls_client.actions.get_list()

        firewalls_client._client.request.assert_called_with(
            url="/firewalls/actions",
            method="GET",
            params={},
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == firewalls_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "set_firewall_rules"

    def test_actions_get_all(self, firewalls_client, response_get_actions):
        firewalls_client._client.request.return_value = response_get_actions
        actions = firewalls_client.actions.get_all()

        firewalls_client._client.request.assert_called_with(
            url="/firewalls/actions",
            method="GET",
            params={"page": 1, "per_page": 50},
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == firewalls_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "set_firewall_rules"



from __future__ import annotations

from .client import BoundNetwork, NetworksClient, NetworksPageResult  # noqa: F401
from .domain import (  # noqa: F401
    CreateNetworkResponse,
    Network,
    NetworkRoute,
    NetworkSubnet,
)



from __future__ import annotations

from .client import (  # noqa: F401
    BoundLoadBalancer,
    LoadBalancersClient,
    LoadBalancersPageResult,
)
from .domain import (  # noqa: F401
    CreateLoadBalancerResponse,
    GetMetricsResponse,
    IPv4Address,
    IPv6Network,
    LoadBalancer,
    LoadBalancerAlgorithm,
    LoadBalancerHealtCheckHttp,
    LoadBalancerHealthCheck,
    LoadBalancerService,
    LoadBalancerServiceHttp,
    LoadBalancerTarget,
    LoadBalancerTargetHealthStatus,
    LoadBalancerTargetIP,
    LoadBalancerTargetLabelSelector,
    PrivateNet,
    PublicNetwork,
)



from __future__ import annotations

import pytest


@pytest.fixture()
def network_response():
    return {
        "network": {
            "id": 1,
            "name": "mynet",
            "created": "2016-01-30T23:50:11+00:00",
            "ip_range": "10.0.0.0/16",
            "subnets": [
                {
                    "type": "cloud",
                    "ip_range": "10.0.1.0/24",
                    "network_zone": "eu-central",
                    "gateway": "10.0.0.1",
                },
                {
                    "type": "vswitch",
                    "ip_range": "10.0.3.0/24",
                    "network_zone": "eu-central",
                    "gateway": "10.0.3.1",
                },
            ],
            "routes": [{"destination": "10.100.1.0/24", "gateway": "10.0.1.1"}],
            "expose_routes_to_vswitch": False,
            "servers": [42],
            "protection": {"delete": False},
            "labels": {},
        }
    }


@pytest.fixture()
def two_networks_response():
    return {
        "networks": [
            {
                "id": 1,
                "name": "mynet",
                "created": "2016-01-30T23:50:11+00:00",
                "ip_range": "10.0.0.0/16",
                "subnets": [
                    {
                        "type": "cloud",
                        "ip_range": "10.0.1.0/24",
                        "network_zone": "eu-central",
                        "gateway": "10.0.0.1",
                    },
                    {
                        "type": "vswitch",
                        "ip_range": "10.0.3.0/24",
                        "network_zone": "eu-central",
                        "gateway": "10.0.3.1",
                    },
                ],
                "routes": [{"destination": "10.100.1.0/24", "gateway": "10.0.1.1"}],
                "expose_routes_to_vswitch": False,
                "servers": [42],
                "protection": {"delete": False},
                "labels": {},
            },
            {
                "id": 2,
                "name": "myanothernet",
                "created": "2016-01-30T23:50:11+00:00",
                "ip_range": "12.0.0.0/8",
                "subnets": [
                    {
                        "type": "cloud",
                        "ip_range": "12.0.1.0/24",
                        "network_zone": "eu-central",
                        "gateway": "12.0.0.1",
                    }
                ],
                "routes": [{"destination": "12.100.1.0/24", "gateway": "12.0.1.1"}],
                "expose_routes_to_vswitch": False,
                "servers": [45],
                "protection": {"delete": False},
                "labels": {},
            },
        ]
    }


@pytest.fixture()
def one_network_response():
    return {
        "networks": [
            {
                "id": 1,
                "name": "mynet",
                "created": "2016-01-30T23:50:11+00:00",
                "ip_range": "10.0.0.0/16",
                "subnets": [
                    {
                        "type": "cloud",
                        "ip_range": "10.0.1.0/24",
                        "network_zone": "eu-central",
                        "gateway": "10.0.0.1",
                    },
                    {
                        "type": "vswitch",
                        "ip_range": "10.0.3.0/24",
                        "network_zone": "eu-central",
                        "gateway": "10.0.3.1",
                    },
                ],
                "routes": [{"destination": "10.100.1.0/24", "gateway": "10.0.1.1"}],
                "expose_routes_to_vswitch": False,
                "servers": [42],
                "protection": {"delete": False},
                "labels": {},
            }
        ]
    }


@pytest.fixture()
def network_create_response():
    return {
        "network": {
            "id": 4711,
            "name": "mynet",
            "ip_range": "10.0.0.0/16",
            "subnets": [
                {
                    "type": "cloud",
                    "ip_range": "10.0.1.0/24",
                    "network_zone": "eu-central",
                    "gateway": "10.0.0.1",
                }
            ],
            "routes": [{"destination": "10.100.1.0/24", "gateway": "10.0.1.1"}],
            "expose_routes_to_vswitch": False,
            "servers": [42],
            "protection": {"delete": False},
            "labels": {},
            "created": "2016-01-30T23:50:00+00:00",
        }
    }


@pytest.fixture()
def network_create_response_with_expose_routes_to_vswitch():
    return {
        "network": {
            "id": 4711,
            "name": "mynet",
            "ip_range": "10.0.0.0/16",
            "subnets": [
                {
                    "type": "cloud",
                    "ip_range": "10.0.1.0/24",
                    "network_zone": "eu-central",
                    "gateway": "10.0.0.1",
                }
            ],
            "routes": [{"destination": "10.100.1.0/24", "gateway": "10.0.1.1"}],
            "expose_routes_to_vswitch": True,
            "servers": [42],
            "protection": {"delete": False},
            "labels": {},
            "created": "2016-01-30T23:50:00+00:00",
        }
    }


@pytest.fixture()
def response_update_network():
    return {
        "network": {
            "id": 4711,
            "name": "new-name",
            "ip_range": "10.0.0.0/16",
            "subnets": [
                {
                    "type": "cloud",
                    "ip_range": "10.0.1.0/24",
                    "network_zone": "eu-central",
                    "gateway": "10.0.0.1",
                }
            ],
            "routes": [{"destination": "10.100.1.0/24", "gateway": "10.0.1.1"}],
            "expose_routes_to_vswitch": True,
            "servers": [42],
            "protection": {"delete": False},
            "labels": {},
            "created": "2016-01-30T23:50:00+00:00",
        }
    }


@pytest.fixture()
def response_get_actions():
    return {
        "actions": [
            {
                "id": 13,
                "command": "add_subnet",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [{"id": 4711, "type": "network"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            }
        ]
    }



from __future__ import annotations

import pytest


@pytest.fixture()
def image_response():
    return {
        "image": {
            "id": 4711,
            "type": "snapshot",
            "status": "available",
            "name": "ubuntu-20.04",
            "description": "Ubuntu 20.04 Standard 64 bit",
            "image_size": 2.3,
            "disk_size": 10,
            "created": "2016-01-30T23:50+00:00",
            "created_from": {"id": 1, "name": "Server"},
            "bound_to": 1,
            "os_flavor": "ubuntu",
            "os_version": "16.04",
            "architecture": "x86",
            "rapid_deploy": False,
            "protection": {"delete": False},
            "deprecated": "2018-02-28T00:00:00+00:00",
            "labels": {},
        }
    }


@pytest.fixture()
def two_images_response():
    return {
        "images": [
            {
                "id": 4711,
                "type": "snapshot",
                "status": "available",
                "name": "ubuntu-20.04",
                "description": "Ubuntu 20.04 Standard 64 bit",
                "image_size": 2.3,
                "disk_size": 10,
                "created": "2016-01-30T23:50+00:00",
                "created_from": {"id": 1, "name": "Server"},
                "bound_to": None,
                "os_flavor": "ubuntu",
                "os_version": "16.04",
                "architecture": "x86",
                "rapid_deploy": False,
                "protection": {"delete": False},
                "deprecated": "2018-02-28T00:00:00+00:00",
                "labels": {},
            },
            {
                "id": 4712,
                "type": "system",
                "status": "available",
                "name": "ubuntu-18.10",
                "description": "Ubuntu 18.10 Standard 64 bit",
                "image_size": 2.3,
                "disk_size": 10,
                "created": "2016-01-30T23:50+00:00",
                "created_from": {"id": 1, "name": "Server"},
                "bound_to": None,
                "os_flavor": "ubuntu",
                "os_version": "16.04",
                "architecture": "x86",
                "rapid_deploy": False,
                "protection": {"delete": False},
                "deprecated": "2018-02-28T00:00:00+00:00",
                "labels": {},
            },
        ]
    }


@pytest.fixture()
def one_images_response():
    return {
        "images": [
            {
                "id": 4711,
                "type": "snapshot",
                "status": "available",
                "name": "ubuntu-20.04",
                "description": "Ubuntu 20.04 Standard 64 bit",
                "image_size": 2.3,
                "disk_size": 10,
                "created": "2016-01-30T23:50+00:00",
                "created_from": {"id": 1, "name": "Server"},
                "bound_to": None,
                "os_flavor": "ubuntu",
                "os_version": "16.04",
                "architecture": "x86",
                "rapid_deploy": False,
                "protection": {"delete": False},
                "deprecated": "2018-02-28T00:00:00+00:00",
                "labels": {},
            }
        ]
    }


@pytest.fixture()
def response_update_image():
    return {
        "image": {
            "id": 4711,
            "type": "snapshot",
            "status": "available",
            "name": None,
            "description": "My new Image description",
            "image_size": 2.3,
            "disk_size": 10,
            "created": "2016-01-30T23:50+00:00",
            "created_from": {"id": 1, "name": "Server"},
            "bound_to": None,
            "os_flavor": "ubuntu",
            "os_version": "16.04",
            "architecture": "arm",
            "rapid_deploy": False,
            "protection": {"delete": False},
            "deprecated": "2018-02-28T00:00:00+00:00",
            "labels": {},
        }
    }


@pytest.fixture()
def response_get_actions():
    return {
        "actions": [
            {
                "id": 13,
                "command": "change_protection",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [{"id": 42, "type": "image"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            }
        ]
    }



from __future__ import annotations

import datetime
from datetime import timezone
from unittest import mock

import pytest

from hcloud.isos import BoundIso, IsosClient


class TestBoundIso:
    @pytest.fixture()
    def bound_iso(self, hetzner_client):
        return BoundIso(client=hetzner_client.isos, data=dict(id=14))

    def test_bound_iso_init(self, iso_response):
        bound_iso = BoundIso(client=mock.MagicMock(), data=iso_response["iso"])

        assert bound_iso.id == 4711
        assert bound_iso.name == "FreeBSD-11.0-RELEASE-amd64-dvd1"
        assert bound_iso.description == "FreeBSD 11.0 x64"
        assert bound_iso.type == "public"
        assert bound_iso.architecture == "x86"
        with pytest.deprecated_call():
            assert bound_iso.deprecated == datetime.datetime(
                2018, 2, 28, 0, 0, tzinfo=timezone.utc
            )
        assert bound_iso.deprecation.announced == datetime.datetime(
            2018, 1, 28, 0, 0, tzinfo=timezone.utc
        )
        assert bound_iso.deprecation.unavailable_after == datetime.datetime(
            2018, 2, 28, 0, 0, tzinfo=timezone.utc
        )


class TestIsosClient:
    @pytest.fixture()
    def isos_client(self):
        return IsosClient(client=mock.MagicMock())

    def test_get_by_id(self, isos_client, iso_response):
        isos_client._client.request.return_value = iso_response
        iso = isos_client.get_by_id(1)
        isos_client._client.request.assert_called_with(url="/isos/1", method="GET")
        assert iso._client is isos_client
        assert iso.id == 4711
        assert iso.name == "FreeBSD-11.0-RELEASE-amd64-dvd1"

    @pytest.mark.parametrize(
        "params",
        [
            {},
            {"name": ""},
            {"name": "FreeBSD-11.0-RELEASE-amd64-dvd1", "page": 1, "per_page": 2},
        ],
    )
    def test_get_list(self, isos_client, two_isos_response, params):
        isos_client._client.request.return_value = two_isos_response
        result = isos_client.get_list(**params)
        isos_client._client.request.assert_called_with(
            url="/isos", method="GET", params=params
        )

        isos = result.isos
        assert result.meta is None

        assert len(isos) == 2

        isos1 = isos[0]
        isos2 = isos[1]

        assert isos1._client is isos_client
        assert isos1.id == 4711
        assert isos1.name == "FreeBSD-11.0-RELEASE-amd64-dvd1"

        assert isos2._client is isos_client
        assert isos2.id == 4712
        assert isos2.name == "FreeBSD-11.0-RELEASE-amd64-dvd1"

    @pytest.mark.parametrize(
        "params", [{}, {"name": "FreeBSD-11.0-RELEASE-amd64-dvd1"}]
    )
    def test_get_all(self, isos_client, two_isos_response, params):
        isos_client._client.request.return_value = two_isos_response
        isos = isos_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        isos_client._client.request.assert_called_with(
            url="/isos", method="GET", params=params
        )

        assert len(isos) == 2

        isos1 = isos[0]
        isos2 = isos[1]

        assert isos1._client is isos_client
        assert isos1.id == 4711
        assert isos1.name == "FreeBSD-11.0-RELEASE-amd64-dvd1"

        assert isos2._client is isos_client
        assert isos2.id == 4712
        assert isos2.name == "FreeBSD-11.0-RELEASE-amd64-dvd1"

    def test_get_by_name(self, isos_client, one_isos_response):
        isos_client._client.request.return_value = one_isos_response
        iso = isos_client.get_by_name("FreeBSD-11.0-RELEASE-amd64-dvd1")

        params = {"name": "FreeBSD-11.0-RELEASE-amd64-dvd1"}

        isos_client._client.request.assert_called_with(
            url="/isos", method="GET", params=params
        )

        assert iso._client is isos_client
        assert iso.id == 4711
        assert iso.name == "FreeBSD-11.0-RELEASE-amd64-dvd1"



from __future__ import annotations

from .client import (  # noqa: F401
    BoundDatacenter,
    DatacentersClient,
    DatacentersPageResult,
)
from .domain import Datacenter, DatacenterServerTypes  # noqa: F401
